/* GNU Chess 5.0 - book.c - book code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
     cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
     lukas@debian.org
*/

// includes

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "common.h"
#include "book.h"
#include "move.h"

// defines

/* The book which BookBuilder() uses */
#define BOOKRUN "book.dat"
/*
 * The last byte of magic_str should be the version
 * number of the format, in case we have to change it.
 *
 * Format 0x01 had an index field which is now gone.
 *
 * Format 0x03 uses an additional size header, which 
 * comes directly after the magic string, and has the
 * number of entries as a big-endian encoded uint32_t
 * number. 
 */

#define MAGIC_LENGTH 5
#define MAXMATCH 100

/*
 * We now allocate memory for the book dynamically,
 * according to the size field in the header of the binary
 * book. However, during book building the following value
 * is used.
 */
#define MAX_DIGEST_BITS 20
#define DIGEST_SIZE(book) (1UL << (book)->digest_bits)
#define DIGEST_MASK(book) (DIGEST_SIZE(book) - 1)
/*
 * Initial hash function, relies on the quality of the lower
 * bits of the 64 bit hash function
 */
#define DIGEST_START(book, i,key)		\
  ((i) = (key) & DIGEST_MASK(book))
/*
 * See if there is already a matching entry
 */
#define DIGEST_MATCH(book,i,the_key)		\
  ((the_key) == (book)->bookpos[i].key)

/*
 * See if the entry is empty
 */
#define DIGEST_EMPTY(book, i) is_empty(book, i)

/*
 * Check for collision
 */
#define DIGEST_COLLISION(book, i,key)		\
  (!DIGEST_MATCH(book, i,key) && !DIGEST_EMPTY(book, i))

/*
 * The next macro is used in the case of hash collisions.
 * We use double hashing with the higher bits of the key.
 * In order to have the shift relatively prime to the hash
 * size, we OR by 1.
 */
#define DIGEST_NEXT(book,i,key)					\
  ((i) = ((i) + (((key) >> (book)->digest_bits) | 1)) & DIGEST_MASK(book))


/*
 * This is considered to be the limit for the hash, I chose
 * 95% because it is Monday... No, I am open for suggestions
 * for the right value, I just don't know.
 */
#define DIGEST_LIMIT(book) (0.95 * DIGEST_SIZE(book))


// constants

//Since the hashcode has changed we change the magic
//static const char magic_str[] = "\x42\x23\x08\x15\x03";

static const char magic_str[] = "GnuCB";
/* Offsets */
static const int wins_off   = 0;
static const int losses_off = 2;
static const int draws_off  = 4;
static const int key_off    = 6;


// types

/*
 * This is the record as it will be written in the binary
 * file in network byte order. HashType is uint64_t. To
 * avoid endianness and padding issues, we do not read or
 * write structs but put the values in an unsigned char array.
 */

typedef unsigned char buf_t[2+2+2+8];


// variables

static char const * bookbin[] = {
   "book.dat",
   "/usr/share/gnuchess/book.dat",
   "/usr/share/games/gnuchess/book.dat",
   "/usr/lib/games/gnuchess/book.dat",
   NULL
};

book_t Book[1];

// functions

static FILE* BookFileOpen(const char *file);
static int   BookLoad(book_t *book);
static int   BookIsLoaded(book_t *book);
static void  BookClear(book_t *book);

void BookAdd (const char *file)
/****************************************************************************
 *
 *  To read a game from a PGN file and store out the hash entries to book.
 *
 ****************************************************************************/
{
    FILE *fp;
    int i;
    int ret;
    int moveno, ngames = 0;
    time_t t1, t2;
    double et;
    int error;
    board_t board[1];
    pgn_iter_t pgn_iter[1];
    game_t game[1];
    book_t book[1];
    //int legacy_filter=false;
    int legacy_filter=true;
    int addtobook[2]={true,true};
    
    et = 0.0;
    t1 = time(NULL);
    
    GameInit(game);
    BookInit(book);
    if (BookBuilderOpen(book) != BOOK_SUCCESS){
	return;
    }
    PGNIterInit(pgn_iter);
    PGNIterStart(pgn_iter,file);
    
    while(PGNIterNext(pgn_iter,game)){
	if(legacy_filter){
	    const char *white_title;
	    const char *black_title;
	    addtobook[white]=false;
	    addtobook[black]=false;
	    white_title=game->white_title;
	    black_title=game->black_title;
	    if (white_title &&
	    	(strcmp(white_title, "GM") == 0 ||
	    	 strcmp(white_title, "IM") == 0 ||
	    	 strcmp(white_title, "FM") == 0)){
	    	addtobook[white]=true;
	    }
	    if (black_title &&
	    	(strcmp(black_title, "GM") == 0 ||
	    	 strcmp(black_title, "IM") == 0 ||
	    	 strcmp(black_title, "FM") == 0)){
	    	addtobook[black]=true;
	    }
	    if(game->white_name && IsTrustedPlayer(game->white_name)){
	    	addtobook[white]=true;
	    }
	    if(game->black_name && IsTrustedPlayer(game->black_name)){
	    	addtobook[black]=true;
	    }
	}
	BoardFromGame(board,game,BOARD_FIRST);
	for(i=game->InitialGameCnt;i<game->RealGameCnt;i++){
	    MoveMake(board,game->boards[i-game->InitialGameCnt].move);
	    if( (addtobook[1^board->side]) && (BookBuilder (book, board,game->result, 1^board->side)!=0)){
	    	break;
	    };
	    //	    MoveMake(board,game->boards[i-game->InitialGameCnt].move);
	}
	ngames++;
	if(ngames%10==0){
	    Output("Games processed: %d\r",ngames);
	}
    }
    
    t2 = time(NULL);
    et += difftime(t2, t1);
    putchar('\n');
    
    /* Handle divide-by-zero problem */
    if (et < 0.5) { et = 1.0; };
    
    Output("Time = %.0f seconds\n", et);
    Output("Games compiled: %d\n",ngames);
    Output("Games per second: %f\n",ngames/et);
    Output("Positions scanned: %d\n",book->newpos+book->existpos);
    Output("Positions per second: %f\n",(book->newpos+book->existpos)/et);
    Output("New & unique added: %d positions\n",book->newpos);
    Output("Duplicates not added: %d positions\n",book->existpos);
    BookBuilderClose(book);
    

}



static int check_magic(FILE *f)
{
  char buf[MAGIC_LENGTH];
  int r;

  r = fread(&buf, 1, MAGIC_LENGTH, f);
  return (r == MAGIC_LENGTH &&
	  memcmp(buf, magic_str, MAGIC_LENGTH) == 0);
}

static int write_magic(FILE *f)
{
  if (MAGIC_LENGTH != fwrite(&magic_str, 1, MAGIC_LENGTH, f)) {
    return BOOK_EIO;
  } else {
    return BOOK_SUCCESS;
  }
}

/* Write and read size information for the binary book */

static int write_size(FILE *f, uint32_t size)
{
  unsigned char sizebuf[4];
  int k;

  for (k = 0; k < 4; k++) {
    sizebuf[k] = (size >> ((3-k)*8)) & 0xff;
  }
  if (1 == fwrite(&sizebuf, sizeof(sizebuf), 1, f)) {
    return BOOK_SUCCESS;
  } else {
    return BOOK_EIO;
  }
}

/* Returns 0 if some read error occurs */

static uint32_t read_size(FILE *rfp)
{
  unsigned char sizebuf[4];
  uint32_t size = 0;
  int k;

  if (1 != fread(&sizebuf, sizeof(sizebuf), 1, rfp)) {
    return 0;
  }
  for (k = 0; k < 4; k++) {
    size = (size << 8) | sizebuf[k];
  }
  return size;
}


static inline int is_empty(book_t *book, uint32_t index)
{
  return 
    book->bookpos[index].key    == 0 &&
    book->bookpos[index].wins   == 0 &&
    book->bookpos[index].draws  == 0 &&
    book->bookpos[index].losses == 0;
}




static void buf_to_book(buf_t buf, book_t *book)
{
  HashType key;
  uint32_t i;

  key = ((uint64_t)buf[key_off] << 56)
    | ((uint64_t)buf[key_off+1] << 48)
    | ((uint64_t)buf[key_off+2] << 40)
    | ((uint64_t)buf[key_off+3] << 32)
    | ((uint64_t)buf[key_off+4] << 24)
    | ((uint64_t)buf[key_off+5] << 16)
    | ((uint64_t)buf[key_off+6] << 8)
    | ((uint64_t)buf[key_off+7]);
  /*
   * This is an infinite loop if the hash is 100% full,
   * but other parts should check that this does not happen.
   */
  for (DIGEST_START(book, i, key);
       DIGEST_COLLISION(book, i, key);
       DIGEST_NEXT(book, i, key))
    /* Skip */
    book->bookhashcollisions++;

  book->bookpos[i].wins   += (buf[wins_off]   << 8) | buf[wins_off  +1];
  book->bookpos[i].draws  += (buf[draws_off]  << 8) | buf[draws_off +1];
  book->bookpos[i].losses += (buf[losses_off] << 8) | buf[losses_off+1];
  book->bookpos[i].key = key;
}

static void book_to_buf(book_t *book, buf_t buf, uint32_t index)
{
  int k;

  for (k=0; k<2; k++) {
    buf[wins_off + k]   = ((book->bookpos[index].wins)   >> ((1-k)*8)) & 0xff;
    buf[draws_off + k]  = ((book->bookpos[index].draws)  >> ((1-k)*8)) & 0xff;
    buf[losses_off + k] = ((book->bookpos[index].losses) >> ((1-k)*8)) & 0xff;
  }
  for (k=0; k<8; k++) {
    buf[key_off + k] = ((book->bookpos[index].key) >> ((7-k)*8)) & 0xff;
  }
}

static int compare(const void *aa, const void *bb)
{
  buf_t buf;
  const leaf *a = aa;
  const leaf *b = bb;
  
  if (b->score > a->score) return(1);
  else if (b->score < a->score) return(-1);
  else return(0);
}

/*
 * Reads an existing binary book from f. The header must
 * already be skipped, when you call this function. The
 * variable digest_bits must be set to the correct value
 * before calling this function. If any book was allocated
 * before, it will be lost.
 */
static int read_book(book_t *book, FILE *rfp)
{
  buf_t buf;
  ASSERT(book->bookpos==NULL);
  ASSERT(book->bookcnt==0);
  ASSERT(book->bookhashcollisions==0);
  book->bookpos = calloc(DIGEST_SIZE(book), sizeof(struct hashtype));
  if (book->bookpos == NULL) {
    return BOOK_ENOMEM;
  }
  while ( 1 == fread(&buf, sizeof(buf), 1, rfp) ) {
    buf_to_book(buf, book);
    book->bookcnt++;
  }
  return BOOK_SUCCESS;
}
  
/* 
 * Return values are defined in book.h
 */

int BookBuilderOpen(book_t *book)
{
  book_t book_tmp_w[1];
  FILE *rfp, *wfp;
  int res;

  rfp=BookFileOpen(BOOKRUN);

  if (rfp != NULL) {
    Output("Opened existing book!\n");
    /*
     * We have to read the size header, but in book building we
     * use the maximum-sized hash table, so we discard the value.
     */
    book->digest_bits = MAX_DIGEST_BITS;
    read_size(rfp);
    res = read_book(book,rfp);
    fclose(rfp);
    if (res != BOOK_SUCCESS) {
      return res;
    }
    Output("Read %d book positions\n", book->bookcnt);
    Output("Got %d hash collisions\n", book->bookhashcollisions);
  } else {
    wfp = fopen(BOOKRUN,"w+b");
    if (wfp == NULL) {
      fprintf(stderr, "Could not create %s file: %s\n",
		 BOOKRUN, strerror(errno));
      return BOOK_EIO;
    }
    if (write_magic(wfp) != BOOK_SUCCESS) {
      fprintf(stderr, "Could not write to %s: %s\n",
	      BOOKRUN, strerror(errno));
      return BOOK_EIO;
    }
    if (fclose(wfp) != 0) {
      fprintf(stderr, "Could not write to %s: %s\n",
	      BOOKRUN, strerror(errno));
      return BOOK_EIO;
    }
    Output("Created new book %s!\n", BOOKRUN);
    rfp = fopen(BOOKRUN, "rb");
    if (rfp == NULL) {
      Output("Could not open %s for reading: %s\n",
	      BOOKRUN, strerror(errno));
      return BOOK_EIO;
    }
    book->digest_bits = MAX_DIGEST_BITS;
    /* We use read_book() here only to allocate memory */
    if (read_book(book,rfp) == BOOK_ENOMEM) {
      return BOOK_ENOMEM;
    }
  }
  return BOOK_SUCCESS;
}

/*
 * Store the position and results of last search
 * if and only if in the first 10 moves of the game.
 * This routine is called after the computer makes its
 * most recent move. Lastly, after the 10th move, on 
 * the 11th move, store the data out to the running file.
 */
/*
 * NOTE: Before you build a book, you have to call
 * BookBuilderOpen() now, after building it BookBuilderClose()
 * in order to actually write the book to disk.
 */

int BookBuilder(book_t *book, board_t *board, short result, uint8_t side)
{
  uint32_t i;
  
  /* Only first BOOKDEPTH moves */
  if (board->GameCnt > BOOKDEPTH+1){ 
    return BOOK_EMIDGAME;
  }
  //  CalcHashKey(Board);
  for (DIGEST_START(book, i, board->HashKey);
       ; 
       DIGEST_NEXT(book, i, board->HashKey)) {
    if (DIGEST_MATCH(book, i, board->HashKey)) {
      book->existpos++;
      break;
    } else if (DIGEST_EMPTY(book, i)) {
      if (book->bookcnt > DIGEST_LIMIT(book))
	return BOOK_EFULL;
      book->bookpos[i].key = board->HashKey;
      book->newpos++;
      book->bookcnt++;
      break;
    } else {
      book->bookhashcollisions++;
    }
  }
  if (side == white) {
    if (result == R_WHITE_WINS)
      book->bookpos[i].wins++;
    else if (result == R_BLACK_WINS)
      book->bookpos[i].losses++;
    else if (result == R_DRAW)
      book->bookpos[i].draws++;
  } else {
    if (result == R_WHITE_WINS)
      book->bookpos[i].losses++;
    else if (result == R_BLACK_WINS)
      book->bookpos[i].wins++;
    else if (result == R_DRAW)
      book->bookpos[i].draws++;
  }
  return BOOK_SUCCESS;
}

int BookBuilderClose(book_t *book)
{
  /*
   * IMHO the following part needs no temporary file.
   * If two gnuchess invocations try to write the same
   * file at the same time, this goes wrong anyway.
   * Please correct me if I am wrong. If you generate
   * a temporary file, try to generate it in some writable
   * directory.
   */
  book_t book_tmp_w[1];
  FILE *wfp;
  buf_t buf;
  unsigned int i;
  int errcode = BOOK_SUCCESS;

  wfp = fopen(BOOKRUN, "wb");
  if (wfp == NULL) {
    errcode = BOOK_EIO;
    goto bailout_noclose;
  }
  if (write_magic(wfp) != BOOK_SUCCESS) {
    errcode = BOOK_EIO;
    goto bailout;
  }
  if (write_size(wfp, book->bookcnt) != BOOK_SUCCESS) {
    errcode = BOOK_EIO;
    goto bailout;
  }
  for (i = 0; i < DIGEST_SIZE(book); i++) {
    if (!is_empty(book,i)) {
      book_to_buf(book,buf,i);
      if (1 != fwrite(&buf, sizeof(buf), 1, wfp)) {
	errcode = BOOK_EIO;
	goto bailout;
      }
    }
  }
  Output("Got %d hash collisions\n", book->bookhashcollisions);

 bailout:
  if (fclose(wfp) != 0) {
    errcode = BOOK_EIO;
  }

 bailout_noclose:
  BookClear(book);

  return errcode;
}


static FILE* BookFileOpen(const char *file){
  FILE* rfp;
   rfp = fopen(file, "rb");
   if (rfp == NULL) return 0;
   OutputConsole("Read opening book (%s)...\n", file);
   if (!check_magic(rfp)) {
     Output( " File %s does not conform to the current format.\n"
	     " Consider rebuilding it.\n\n",
	     file);
     fclose(rfp);
     rfp=NULL;
   }
   return rfp;
}

void BookInit(book_t *book){
  book->filename=NULL;
  book->bookpos=NULL;
  book->bookcnt=0;
  book->digest_bits=0;
  book->bookhashcollisions=0;
  book->bookfirstlast=3;
  book->newpos=0;
  book->existpos=0;
}


void BookClear(book_t *book){
  if(book->bookpos){
    free(book->bookpos);
  }
  BookInit(book);
}


void BookSetFile(book_t *book, const char *filename){
  BookClear(book);
  my_string_set(&book->filename,filename);
}



int BookLoad(book_t *book){
  FILE *rfp;
  int booksize,res;
  rfp=BookFileOpen(book->filename);
  if(!rfp) return BOOK_ENOBOOK;
  OutputConsole("Loading book from %s.\n", book->filename);
  /*
   * The factor 1.06 is the minimum additional amount of
   * free space in the hash
   */
  booksize = read_size(rfp) * 1.06;
  for (book->digest_bits = 1; booksize; booksize >>= 1) {
    book->digest_bits++;
  }
  // allocates memory
  res = read_book(book,rfp);
  fclose(rfp);
  if (res != BOOK_SUCCESS) {
    return res;
  }
  OutputConsole("%d hash collisions... \n", book->bookhashcollisions);
  return res;
}


const char* BookFind(){
  FILE *rfp;
  char const * const *booktry;


  for (booktry = bookbin; *booktry ; booktry++) {
    OutputConsole("Looking for opening book in %s...\n", *booktry);
    rfp=BookFileOpen(*booktry);
    if(rfp==NULL){
      continue;
    }else{
      break;
    }
  }
  if (rfp == NULL) {
    /* If appropriate print error */
    Output(" No book found.\n\n");
    return NULL;
  }
  fclose(rfp);
  return *booktry;
}

int BookIsLoaded(book_t *book){
  return book->bookpos!=NULL;
}


/*
 * Return values are defined in book.h
 * NOTE: This function now returns BOOK_SUCCESS on
 * success, which is 0. So all the possible callers
 * should adjust to this. (At present it is only called
 * in iterate.c.) It used to return 1 on success before.
 */

int BookQuery(book_t *book, board_t *board, movelist_t *movelist)
{
  int i,j,k,icnt, mcnt, found, tot, maxdistribution;
  int ret;
  int matches[MAXMATCH] ;
  leaf m[MAXMOVES];
  leaf pref[MAXMOVES];
  HashType posshash[MAXMOVES];
  char SANmv[SANSZ];
  FILE *rfp;
  undo_t undo[1];
  int BKquery=0;

  struct {
    uint16_t wins;
    uint16_t losses;
    uint16_t draws;
  } r[MAXMOVES];
  leaf *p;
  short side,xside,temp;
  uint32_t booksize;
  int res;

  movelist->gencnt=0;

  if(!BookIsLoaded(book)){
    if((ret=BookLoad(book))!=BOOK_SUCCESS){
      return ret;
    }
  }
  mcnt = -1;
  icnt= 0;
  GenMoves(board, movelist);
  for (p = movelist->moves; p < movelist->moves+movelist->gencnt; p++) {
    MoveMakeS(board, &p->move, undo);
    m[icnt].move = p->move;
    posshash[icnt] = board->HashKey;
    icnt++;
    MoveUnMakeS(board, undo);
  }
  // I don't understand the next code!
  for (i = 0; i < icnt; i++) {
    for (DIGEST_START(book,j,posshash[i]);
	 !DIGEST_EMPTY(book, j);
	 DIGEST_NEXT(book,j, posshash[i])) {
      if (DIGEST_MATCH(book, j, posshash[i])) {
	found = 0;
	for (k = 0; k < mcnt; k++) 
	  if (matches[k] == i) {
	    found = 1;
	    break;
	  }
	/* Position must have at least some wins to be played by book */
	if (!found) {
	  matches[++mcnt] = i;
	  pref[mcnt].move = m[matches[mcnt]].move;
	  r[i].losses = book->bookpos[j].losses;
	  r[i].wins = book->bookpos[j].wins;
	  r[i].draws = book->bookpos[j].draws;
	
	  /* by percent score starting from this book position */
	  
	  pref[mcnt].score = m[i].score = 
	    100*(r[i].wins+(r[i].draws/2))/
	    (MAX(r[i].wins+r[i].losses+r[i].draws,1)) + r[i].wins/2;
	  
	}
	if (mcnt >= MAXMATCH) {
	  OutputConsole(" Too many matches in book.\n\n");
	  goto fini;
	}
	break; /* If we found a match, the following positions can not match */
      }
    }
  }

fini:  
  for(i=0;i<=mcnt;i++){
    movelist->moves[i].move=pref[i].move;
    movelist->moves[i].score=pref[i].score;
  }
  movelist->gencnt=mcnt+1;
  return BOOK_SUCCESS;

}
