/* GNU Chess 5.0 - common.h - common symbolic definitions
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
     cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
*/

#ifndef COMMON_H
#define COMMON_H


#include <setjmp.h>
#include <config.h>

#ifndef __GNUC__
# define __attribute__(x)
#endif

/*
 * Include "uint64_t" and similar types using the ac_need_stdint_h ac macro
 */

#include "GCint.h"

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <ctype.h>


 /*
  * Define macro for declaring 64bit constants for compilers not using ULL
  */

#ifdef _MSC_VER
   #define ULL(x) ((uint64_t)(x))
#else
   #define ULL(x) x ## ULL
#endif


// [VdB] Some elementary functions borrowed from Polyglot. 
// Written by Fabien Letouzey.

void my_string_set(const char ** variable, const char string[]);
int  my_string_case_equal(const char string_1[], const char string_2[]);
int  my_string_equal(const char string_1[], const char string_2[]);
void my_string_clear(const char ** variable);


// [VdB] Inifile handling functions. Borrowed from Polyglot but written by me.

// defines

#define IniEntriesNb 256

// types

typedef struct {
    const char *section;
    const char *name;
    const char *value;
    const char *comment;
} ini_entry_t;

typedef struct {
    ini_entry_t entries[IniEntriesNb];
    int index;
    int iter;
} ini_t;

typedef enum {
    SYNTAX_ERROR,
    EMPTY_LINE,
    NAME_VALUE,
    EMPTY_VALUE,
    SECTION
} line_type_t;

// functions

extern void         ini_init              (ini_t *ini);
extern void         ini_clear             (ini_t *ini);
extern void         ini_copy              (ini_t *dst, ini_t *src);
extern int          ini_parse             (ini_t *ini, const char *filename);
extern void         ini_disp              (ini_t *ini);
extern void         ini_insert            (ini_t *ini, ini_entry_t *entry);
extern void         ini_insert_ex         (ini_t *ini,
                                           const char *section,
                                           const char *name,
                                           const char *value);
extern void         ini_start_iter        (ini_t *ini);
extern ini_entry_t *ini_next              (ini_t *ini);
extern ini_entry_t *ini_find              (ini_t *ini,
                                           const char *section,
                                           const char *name);
extern line_type_t ini_line_parse         (const char *line,
                                           char *section,
                                           char *name,
                                           char *value);



#define TOKENS 3

extern char *token[TOKENS];


/*
 * BitBoard is a key data type.  It is a 64-bit value, in which each
 * bit represents a square on the board as defined by "enum Square".
 * Thus, bit position 0 represents a fact about square a1, and bit
 * position 63 represents a fact about square h8.  For example, the
 * bitboard representing "white rook" positions will have a bit set
 * for every position occupied on the board by a white rook.
 */

typedef uint32_t KeyType;

#define MAXPLYDEPTH   250
//#define MAXPLYDEPTH   3 // buffer overflow debug
#define MAXMOVES      200
#define MAXGAMEDEPTH  600
#define MAXFEN        600
#define MAXPV          30
#define SANSZ 8 /* longest move is "exf8=Q+" so 7+1 */

#include "bitboard.h"
#include "hash.h"

typedef struct {
    int TCinc;
    float TCTime;
    int TCMove;
    int SearchDepth;
    //  int MoveLimit[2];
    float TimeLimit[2];
    int NPS;
    int ST;
    int NextPeriod;
    float SearchTime;
    float maxtime;
    int computer;
    int computerplays; // during pondering this is 
                       // not the same as "computer"
} searchparams_t;

typedef struct {
    int console;
    int TCinc;
    float TCTime;
    int TCMove;
    int SearchDepth;
    float TimeLimit[2];
    int NPS;
    int ST;
    int NextPeriod;
    int computer;
    const char *name;
    game_t game[1];
    unsigned int state_flags;
} protocol_t;

extern unsigned long TotalPawnHashCnt;
extern unsigned long GoodPawnHashCnt;
extern unsigned long TotalGetHashCnt;
extern unsigned long GoodGetHashCnt;
extern unsigned long TotalPutHashCnt;
extern unsigned long CollHashCnt;




typedef struct
{
    unsigned long NodeCnt;
    unsigned long QuiesCnt;
    unsigned long EvalCnt;
    unsigned long EvalCall;
    unsigned long OneRepCnt;
    unsigned long ChkExtCnt;
    unsigned long RcpExtCnt;
    unsigned long PawnExtCnt;
    unsigned long HorzExtCnt;
    unsigned long ThrtExtCnt;
    unsigned long KingExtCnt;;
    unsigned long NullCutCnt;
    unsigned long FutlCutCnt;
    unsigned long LMRRedCnt;
    unsigned long RazrCutCnt;
    unsigned long GenCnt;
    int RootPV;
    int RootPVScore;
    float ElapsedTime;

} searchstats_t;


/* leaf describes a leaf-level analysis result */

typedef struct
{
   int move;   /* the move that produced this particular board */
   int score;  /* the scored value of this leaf */
} leaf;


#include "movelist.h"
#include "moveseq.h"
#include "timer.h"

typedef struct {
  int phase;
  movelist_t movelist[1];
  movelist_t captures[1];
  leaf killer1[1];
  leaf killer2[1];
  leaf hashmv[1];
  leaf *p;
} pickstate_t;


#define SCORE_INIT 0
#define SCORE_FAIL_LOW 1
#define SCORE_FAIL_HIGH 2
#define SCORE_EXACT 3
#define SCORE_PV 4


typedef struct{

    /* for timeouts */
    jmp_buf rootpointer;
    
    int RootPawns;
    int RootPieces;
    int RootMaterial;
    int RootDepth;
    int RootScore;
    int RootAlpha;
    int RootBeta;
    int RootState;
    moveseq_t RootPVSeq[1];

    /* periodic updates */
    int  RootMoves;
    int  RootMovesRemaining;
    char RootMoveCurrent[SANSZ];
    
    timer__t timer[1];
    unsigned long history[2][4096];
    int lazyscore[2];
    int maxposnscore[2];
    
    int killer1[MAXPLYDEPTH];
    int killer2[MAXPLYDEPTH];
    int ChkCnt[MAXPLYDEPTH]; /* counts the number of checks extensions up to now */
    short InChk[MAXPLYDEPTH];
    short KingThrt[2][MAXPLYDEPTH];
    int Hashmv[MAXPLYDEPTH];
    //    int moves[MAXPLYDEPTH];
    moveseq_t moveseq[1];
} searchdata_t;



typedef struct {
  FILE *fp;
}   epd_iter_t;

typedef struct {
  FILE *fp;
  void* scanner;
}   pgn_iter_t;


typedef struct
{
   KeyType pkey;  
   BitBoard passed;
   BitBoard weaked;
   int score;
   int phase;
} PawnSlot;


/*  MACRO definitions */

#define DEFAULT_NEWLINE_PREFIX "DEBUG: "

#ifndef MAX
#define MAX(a,b)     ((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b)     ((a) < (b) ? (a) : (b))
#endif
#define SET(a,b)                     \
  do {                               \
    (a) |= (b);                      \
    Log("Set   0x%x\n", (b)); \
  } while (0)
#define CLEAR(a,b)                   \
  do {                               \
    (a) &= ~(b);                     \
    Log("Clear 0x%x\n", (b)); \
  } while (0)

/* Draw score can be used to penalise draws if desired */
#define	DRAWSCORE   0 
#define MATERIAL(board,side)     ((board)->material[(side)] - (board)->material[1^(side)])
#define PHASEINIT(board)	     (MAX(0,(8 - ((board)->material[white]+(board)->material[black]) / 1150)))
#define PHASE(board)	     (board->phase)
#define KEY(a)	     (a >> 32) 

#define SideToMove(GameCnt) ((GameCnt)%2==0?white:black)

/*  Attack MACROS */

#define BishopAttack(board,sq)						\
  (bitboards.Bishop45Atak[sq][((board)->blockerr45 >> Shift45[sq]) & Mask45[sq]] | \
   bitboards.Bishop315Atak[sq][((board)->blockerr315 >> Shift315[sq]) & Mask315[sq]])
#define RookAttack(board,sq)					  \
  (bitboards.Rook00Atak[sq][((board)->blocker >> Shift00[sq]) & 0xFF] |		\
   bitboards.Rook90Atak[sq][((board)->blockerr90 >> Shift90[sq]) & 0xFF])
#define QueenAttack(board,sq)			\
  (BishopAttack(board,sq) | RookAttack(board,sq))


/*  Some bit macros  */

/*
 * gcc 2.95.4 completely screws up the macros with lookup tables 
 * with -O2 on PPC, maybe this check has to be refined. (I don't know
 * whether other architectures also suffer from this gcc bug.) However,
 * with gcc 3.0, the lookup tables are _much_ faster than this direct
 * calculation.
 */
#if defined(__GNUC__) && defined(__PPC__) && __GNUC__ < 3
#  define SETBIT(b,i)   ((b) |=  ((ULL(1)<<63)>>(i)))
#  define CLEARBIT(b,i) ((b) &= ~((ULL(1)<<63)>>(i)))
#else
#  define SETBIT(b,i)   ((b) |= bitboards.BitPosArray[i])
#  define CLEARBIT(b,i) ((b) &= bitboards.NotBitPosArray[i])
#endif

#define RANK(i) ((i) >> 3)
#define ROW(i) ((i) & 7)
#define SQUARE(file,rank) ((file+(rank<<3)))

#define trailz(b) (leadz ((b) & ((~b) + 1)))


#define white  0
#define black  1
#define false  0
#define true   1
#define ks 0
#define qs 1
#define INFINITY  99999
#define MATE	  9999
#define MATESCORE(a)	((a) > MATE-255  || (a) < -MATE+255)


/*  Game flags */
#define QUIT    0x0001
//#define TESTT   0x0002
#define THINK   0x0004
#define MANUAL  0x0008
//#define TIMEOUT 0x0010
//#define SPARE1  0x0020
//#define ENDED   0x0040
//#define USEHASH 0x0080
  //#define SOLVE   0x0100
//#define USENULL 0x0200
//#define XBOARD  0x0400
//#define TIMECTL 0x0800
#define POST    0x1000
#define PONDER  0x2000 /* We are in pondering (during search) */
#define HARD    0x4000 /* Pondering is turned on */
#define ANALYZE 0x8000 /* In ANALYZE mode */

/*  Node types  */ 
#define PV  0
#define ALL 1
#define CUT 2



/*  The various phases during move selection  */
#define PICKHASH    1
#define PICKGEN1    2
#define PICKCAPT    3
#define PICKKILL1   4
#define PICKKILL2   5
#define PICKGEN2    6
#define PICKHIST    7
#define PICKREST    8
#define PICKCOUNTER 9

  //#define MAXTREEDEPTH  2000

/* 
   Smaller HASHSLOT defaults 20011017 to improve blitz play
   and make it easier to run on old machines
*/
#define PAWNSLOTS 512

#define StartFen "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"

extern short distance[64][64];
extern short taxicab[64][64];
extern unsigned char lzArray[65536];
extern short Shift00[64];
extern short Shift90[64];
extern short Shift45[64];
extern short Shift315[64];
extern short directions[64][64];
extern unsigned char BitCount[65536];




// XXX Remaining troublesome globals.

extern volatile unsigned int flags;
extern unsigned int preanalyze_flags;
extern unsigned HashSize;
extern unsigned long TTHashMask;
extern unsigned long PHashMask;
extern  BitBoard Ataks[2][7];
extern  BitBoard pinned;
extern  BitBoard pieces[2];

extern epd_iter_t EPDIter[1];
extern pgn_iter_t PGNIter[1];
extern protocol_t Protocol[1];


extern HashType hashcode[2][7][64];
extern HashType ephash[64];
extern HashType WKCastlehash;
extern HashType WQCastlehash;
extern HashType BKCastlehash;
extern HashType BQCastlehash;
extern HashType Sidehash;
extern PawnSlot *PawnTab[2];

extern uint16_t trans_age;

extern int slider[8];
extern int Value[7];
extern int range[8];
extern int ptype[2];

extern char algbr[64][3];
extern char algbrfile[9];
extern char algbrrank[9];

extern char notation[8];
extern char lnotation[8];
extern int r90[64];
extern int r45[64];
extern int r315[64];
extern int Mask45[64];
extern int Mask315[64];

extern int rank6[2];
extern int rank7[2];
extern int rank8[2];

extern char *progname;
extern FILE *ofp;

extern searchparams_t SearchParams[1];
extern int n; 		/* Last mobility returned by CTL */
extern int newpos, existpos;		/* For book statistics */

enum Piece { empty, pawn, knight, bishop, rook, queen, king, bpawn };

enum Square { A1, B1, C1, D1, E1, F1, G1, H1,
	      A2, B2, C2, D2, E2, F2, G2, H2,
	      A3, B3, C3, D3, E3, F3, G3, H3,
	      A4, B4, C4, D4, E4, F4, G4, H4,
	      A5, B5, C5, D5, E5, F5, G5, H5,
	      A6, B6, C6, D6, E6, F6, G6, H6,
	      A7, B7, C7, D7, E7, F7, G7, H7,
	      A8, B8, C8, D8, E8, F8, G8, H8 };

enum File { A_FILE, B_FILE, C_FILE, D_FILE, E_FILE, F_FILE, G_FILE, H_FILE };

/****************************************************************************
 *
 *  The various function prototypes.  They are group into the *.c files
 *  in which they are defined.
 *
 ****************************************************************************/

/*
 * Explanation of the #ifdef NO_INLINE conditionals:
 *
 * Define NO_INLINE only if you really must, implementations will be
 * provided by the corresponding *.c files. The better solution is to
 * not define it, in which case inlines.h will be included which
 * provides static inline version of these functions.
 */

/*  The initialization routines  */
void Initialize (void);
void InitLzArray (void);
void InitBitCount (void);
void ProtocolInit(protocol_t *protocol);

void SearchParamsInit(searchparams_t *searchparams);

#include "option.h"
void SearchParamsCreate(searchparams_t *searchparams, 
			protocol_t *protocol,
			option_list_t *option);

#include "board.h"
void SearchDataInit(searchdata_t *searchdata, board_t *board);
void PickStateInit(pickstate_t *pickstate);
void InitVars (void);
void InitHashCode (void);
void InitHashTable (void);
void CalcHashSize (int);
void InitInput (void);

/* Cleanup routines */
void CleanupInput(void);


 
/*  The move generation routines  */
void GenMoves (board_t *board, 
	       movelist_t *movelist);
void GenPseudoCaptures (board_t *board, 
		  movelist_t *movelist);
void GenPseudoNonCaptures (board_t *board, 
		     movelist_t *movelist);
void GenPseudoCheckEscapes (board_t *board, 
		      movelist_t *movelist);
void FilterIllegalMoves (board_t *board, 
			 movelist_t *movelist); 

/*  The move routines  */

leaf *IsInMoveList (movelist_t *movelist,  int f, int t, char piece);

/*  The attack routines  */
short SqAtakd (board_t *board, short sq, short side);
void GenAtaks (board_t *board);
BitBoard AttackTo (board_t *board, int sq, int side);
BitBoard AttackXTo (board_t *board, int sq, int side);
BitBoard AttackFrom (board_t *board, int sq, int piece, int side);
BitBoard AttackXFrom (board_t *board, int sq, int side);
int PinnedOnKing (board_t *board, int sq, int side);
void FindPins (board_t *board, BitBoard *pin);
int MateScan (board_t *board, int side);

/*  The swap routines  */
int SwapOff (board_t *board, int move);
void AddXrayPiece (board_t *board, 
		   int t, 
		   int sq, 
		   int side, 
		   BitBoard *b, 
		   BitBoard *c);

/*  The EPD routines  */
int ParseEPD (board_t *board, const char *fen, char *bm, char *id);
void SaveEPD (board_t *board, char *file);
int EPDIterStart(epd_iter_t *epd_iter, char *file);
void EPDIterClose(epd_iter_t *epd_iter);
int EPDIterNext(epd_iter_t *epd_iter, board_t *board, char *id, char *bm);

/* Error codes for ParseEPD */
enum {
   EPD_SUCCESS,
   EPD_ERROR
};


void UpdateFriends (board_t *board);
void UpdateCBoard (board_t *board);
void UpdateMvboard (board_t *board);
void UpdatePhase (board_t *board);
void EndSearch (int);
short ValidateBoard (board_t *board);

/*  PGN routines  */
void PGNSaveToFile (const char *file, game_t *game, const char *resultstr);
void PGNReadFromFile (game_t *game, const char *file);
int IsTrustedPlayer(const char *name);


/*  The evaluation routines  */
int ScoreP (board_t *board, searchparams_t *searchparams, short side);
int ScoreN (board_t *board, searchparams_t *searchparams,  short side);
int ScoreB (board_t *board, searchparams_t *searchparams, short side);
int ScoreR (board_t *board, searchparams_t *searchparams, short side);
int ScoreQ (board_t *board, searchparams_t *searchparams, short side);
int ScoreK (board_t *board, searchparams_t *searchparams, short side);
int ScoreDev (board_t *board, searchparams_t *searchparams, short side);
int Evaluate (board_t *board, 
	      searchparams_t *searchparams, 
	      searchdata_t *searchdata, 
	      searchstats_t *searchstats, 
	      int alpha, 
	      int beta);
short EvaluateDraw (board_t *board);

/*  Hung routines  */
int EvalHung (board_t *board, int side);

/*  The search routines  */


void Iterate (board_t *board, 
	      searchparams_t *searchparams, 
	      searchstats_t *searchstats);
int Search (board_t *board, 
	    searchparams_t *searchparams, 
	    searchdata_t *searchdata, 
	    searchstats_t *searchstats,
	    uint8_t ply, 
	    short depth, 
	    int alpha, 
	    int beta, 
	    short nodetype, 
	    short parentnodetype,
	    moveseq_t *pv);
void SearchRoot (board_t *board, 
		 searchparams_t *searchparams, 
		 searchdata_t *searchdata, 
		 searchstats_t *searchstats,
		 movelist_t *movelist);
int Quiesce (board_t *board, 
	     searchparams_t *searchparams,
	     searchdata_t *searchdata, 
	     searchstats_t *searchstats, 
	     uint8_t ply, 
	     int alpha, 
	     int beta);
short Repeat (board_t *board);
void ShowLine (board_t *board, 
	       searchparams_t *searchparams, 
	       searchdata_t *searchdata, 
	       searchstats_t *searchstats,
	       int c);

//double GetElapsed (searchparams_t *searchparams, 
//		   searchdata_t *searchdata,
//		   searchstats_t *searchstats);

void PTClear (void);

/*  Sorting routines  */
void SortCaptures (board_t *board, 
		   movelist_t *movelist);
void SortMoves (board_t *board, 
		searchdata_t *searchdata, 
		movelist_t *movelist,
		int ply);
void SortRoot (board_t *board, movelist_t *movelist);
int PhasePick (board_t *board, 
	       searchdata_t *searchdata, 
	       pickstate_t *pickstate,
	       leaf **p1, 
	       int ply);
int PhasePick1 (board_t *board, 
		searchdata_t *searchdata, 
		pickstate_t *pickstate,
		leaf **p1, 
		int ply);

/*  Some output routines */
void ShowSmallBoard (board_t *board);
void ShowBitBoard (BitBoard *);
void ShowCBoard (board_t *board);
void ShowMvboard (board_t *board);

void ShowGame (game_t *game);
void ShowTime (void);

/*  Random numbers routines */
uint32_t Rand32 (void);
HashType Rand64 (void);

/*  Solver routines  */
void Solve (char *file, searchparams_t *searchparams);

/*  Test routines  */
void TestMoveGenSpeed (void);
void TestNonCaptureGenSpeed (void);
void TestCaptureGenSpeed (void);
void TestMoveList (void);
void TestNonCaptureList (void);
void TestCaptureList (void);
void TestEvalSpeed (void);
void TestEval (void);

/* Player database */
void DBSortPlayer (const char *style);
void DBListPlayer (const char *style); 	
void DBReadPlayer (void);	
void DBWritePlayer (void);
int DBSearchPlayer (const char *player);
void DBUpdatePlayer (const char *player, const char *resultstr);
void DBTest (void);

/* Input thread and thread function */
#include <pthread.h>
//extern pthread_t input_thread;
void *input_func(void *);

/*
 * Status variable used by the input thread to signal
 * pending input. Thought about using flags for this
 * but this can be refined and is conceptually different
 * from flags.
 */
enum {
  INPUT_NONE,
  INPUT_AVAILABLE
};
extern volatile int input_status;

/*
 * Function to wake up the input thread, should be called after
 * input has been parsed.
 */
void input_wakeup(void);

/* Wait for input. */

void wait_for_input(void);

/*
 * Input routine, initialized to one of the specific
 * input routines. The given argument is the prompt.
 */
void (*getline_) (char *);

// [VdB] Made this bigger

#define MAXSTR 4096
extern char inputstr[MAXSTR];

/* Input parser */
void parse_input(void);

/* Pondering */
void ponder(board_t *board, 
	    searchparams_t *searchparams, 
	    searchstats_t *searchstats);

/*  The command subroutines */
void ShowCmd (char *);
void BookCmd (char *);
void TestCmd (char *);

/* Commands from the input engine */
void cmd_accepted(void);
void cmd_activate(void); 
void cmd_analyze(void);
void cmd_bk(void);
void cmd_black(void);
void cmd_book(void);
void cmd_computer(void);
void cmd_depth(void);
void cmd_draw(void);
void cmd_easy(void);
void cmd_edit(void); 
void cmd_epd(void);
void cmd_exit(void);
void cmd_force(void);
void cmd_go(void);
void cmd_hard(void);
void cmd_hash(void);
void cmd_hashsize(void);
void cmd_help (void);
void cmd_hint(void);
void cmd_level(void);
void cmd_list(void);
void cmd_load(void);
void cmd_manual(void);
void cmd_movenow(void);
void cmd_name(void);
void cmd_new(void);
void cmd_nopost(void);
void cmd_null(void);
void cmd_otim(void);
void cmd_pgnload(void);
void cmd_pgnsave(void);
void cmd_ping(void);
void cmd_post(void);
void cmd_protover(void);
void cmd_quit(void);
void cmd_random(void);
void cmd_rating(void);
void cmd_rejected(void);
void cmd_remove(void);
void cmd_result(void);
void cmd_save(void);
void cmd_setboard(void);
void cmd_show (void);
void cmd_solve(void);
void cmd_st(void); 
void cmd_switch(void);
void cmd_test (void);
void cmd_time(void);
void cmd_undo(void);
void cmd_usage(void);
void cmd_variant(void);
void cmd_version(void);
void cmd_white(void);
void cmd_xboard(void);

/*
 * Define NO_INLINE only if you really must, implementations will be
 * provided by the corresponding *.c files. The better solution is to
 * not define it, in which case inlines.h will be included which
 * provides static inline version of these functions.
 */

/*  Some utility routines  */
#ifdef NO_INLINE
unsigned char leadz (BitBoard);
unsigned char nbits (BitBoard);
#else
# include "inlines.h"
#endif

/* from polyglot */

#undef ASSERT
#ifdef DEBUG
#include <assert.h>
#  define ASSERT(a) { if(!(a)) {Log("file \"%s\", line %d, assertion \"" #a "\" failed\n",__FILE__,__LINE__); assert(a);}}
#else
#  define ASSERT(a)
#endif


// [VdB] Output functions

void Output(char *fmt, ...);
void OutputNXB(char *fmt, ...);

#endif /* !COMMON_H */
