/* GNU Chess 5.0 - init.c - initialization of variables code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
     cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>

#include "common.h"
#include "version.h"
#include "lexpgn.h"
#include "book.h"
#include "bitboard.h"
#include "log.h"
#include "option.h"
#include "ttable.h"

void Initialize (void)
/**************************************************************************
 *
 *  The main initialization driver.
 *
 **************************************************************************/
{
  const char *bookfile;

   /* Initialize random number generator */
   srand((unsigned int) time(NULL));

   /* initialize control flags */
   flags = ULL(0);


  InitLzArray ();
  BitBoardInit();
  HashKeyInit ();

     
  OptionInit(Option);
  SearchInit(Option);

  LogInit(LogDefault);
  LogSetFile(LogDefault,OptionGet(Option,"LogFile"));
  LogSetState(LogDefault,OptionGetBool(Option,"Log"));
  LogSetTimeStamp(LogDefault,true);
  LogSetNewlinePrefix(LogDefault,DEFAULT_NEWLINE_PREFIX);
  Log("\n\nOPENING LOGFILE\n***************\n\n");
  ProtocolInit(Protocol);

  TTableInit(TTable);
  TTableSetSize(TTable,HASHSLOTS);

  InitHashTable ();
  EPDIterInit(EPDIter);
  PGNIterInit(PGNIter);
  InitInput (); 
  BookInit(Book);
  if(my_string_equal(OptionGet(Option,"BookFile"),"<empty>")){
    bookfile=BookFind();
    if(bookfile){
      OptionSet(Option,"BookFile",bookfile);
    }else{
      OptionSet(Option,"BookFile","<empty>");
    }
  }
  BookSetFile(Book,OptionGet(Option,"BookFile"));


  GameNew(Protocol->game);



}

void ProtocolInit(protocol_t *protocol){
    char *name;
    protocol->console=isatty(fileno(stdout));
    protocol->TCinc=0;
    protocol->TCMove=40;
    protocol->TCTime=5.0;
    protocol->SearchDepth=0;
    protocol->TimeLimit[white]=60*protocol->TCTime;
    protocol->TimeLimit[black]=60*protocol->TCTime;
    protocol->NPS=0;
    protocol->ST=0;
    protocol->NextPeriod=2*protocol->TCMove;
    protocol->computer=black;
    name=getenv("LOGNAME");
    if(name){
	protocol->name=strdup(name);
    }else{
	protocol->name=strdup("Unknown");
    }
    GameInit(protocol->game);
    protocol->state_flags=0;

}


void SearchParamsCreate(searchparams_t *searchparams, 
			protocol_t *protocol, 
			option_list_t *option){
    searchparams->TCinc=protocol->TCinc;
    searchparams->TCTime=protocol->TCTime;
    searchparams->TCMove=protocol->TCMove;
    searchparams->SearchDepth=protocol->SearchDepth;
    searchparams->TimeLimit[white]=protocol->TimeLimit[white];
    searchparams->TimeLimit[black]=protocol->TimeLimit[black];
    searchparams->NPS=protocol->NPS;
    searchparams->ST=protocol->ST;
    searchparams->NextPeriod=protocol->NextPeriod;
    searchparams->computer=protocol->computer;
    
    searchparams->SearchTime=0.0;
    searchparams->maxtime=0.0;
    searchparams->computerplays=black; // random choice
}


void InitHashTable(void){
    int size;
    /*
     * Here realloc() is O.K., though not very useful, because PAWNSLOTS
     * is constant anyway.
     */
    PawnTab[0] = (PawnSlot *) realloc (PawnTab[0], PAWNSLOTS * sizeof 
				       (PawnSlot));
    PawnTab[1] = (PawnSlot *) realloc (PawnTab[1], PAWNSLOTS * sizeof 
				       (PawnSlot));
    if (PawnTab[0] == NULL || PawnTab[1] == NULL) {
	printf ("Not enough memory for pawn table, goodbye.\n");
	exit(EXIT_FAILURE);
    } else {
	size = (PAWNSLOTS * 2 * sizeof (PawnSlot)) >> 10;
    }
    OutputConsole("Pawn hash table: Entries=%dK Size=%dK\n",
		  PAWNSLOTS >> 10, size);

}


#define NBITS 16

void InitLzArray (void)
/***************************************************************************
 *
 *  The lzArray is created.  This array is used when the position
 *  of the leading non-zero bit is required.  The convention used
 *  is that the leftmost bit is considered as position 0 and the
 *  rightmost bit position 63.
 *
 ***************************************************************************/
{
   int i, j, s, n;

   s = n = 1;
   for (i = 0; i < NBITS; i++)
   {
      for (j = s; j < s + n; j++)
         lzArray[j] = NBITS - 1 - i;
      s += n;
      n += n;
   }
}









void InitBitCount (void)
/**************************************************************************
 *
 *  The BitCount array returns the no. of bits present in the 16 bit
 *  input argument.  This is use for counting the number of bits set
 *  in a BitBoard (e.g. for mobility count).
 *
 **************************************************************************/
{
   int i, j, n;
 
   BitCount[0] = 0;
   BitCount[1] = 1; 
   i = 1;
   for (n = 2; n <= 16; n++)
   {
      i <<= 1;
      for (j = i; j <= i + (i-1); j++)  
         BitCount[j] = 1 + BitCount[j - i]; 
   }
} 



void BoardInit(board_t *board){


  /* Initialize board */
  int i;
  //  memset (board, 0, sizeof (board_t));
  // for (i = 8; i < 16; i++){
  //    SETBIT (board->b[white][pawn], i);
  // }
  // SETBIT (board->b[white][rook], 0);
  // SETBIT (board->b[white][knight], 1);
  // SETBIT (board->b[white][bishop], 2);
  // SETBIT (board->b[white][queen], 3);
  // SETBIT (board->b[white][king], 4);
  // SETBIT (board->b[white][bishop], 5);
  // SETBIT (board->b[white][knight], 6);
  // SETBIT (board->b[white][rook], 7);
  // for (i = 48; i < 56; i++){
  //   SETBIT (board->b[black][pawn], i);
  // }
  // SETBIT (board->b[black][rook], 56);
  // SETBIT (board->b[black][knight], 57);
  // SETBIT (board->b[black][bishop], 58);
  // SETBIT (board->b[black][queen], 59);
  // SETBIT (board->b[black][king], 60);
  // SETBIT (board->b[black][bishop], 61);
  // SETBIT (board->b[black][knight], 62);
  // SETBIT (board->b[black][rook], 63);
   //   board->flag |= (WCASTLE | BCASTLE);
  //   board->ep = -1;
  //   board->king[white] = E1;
  // board->king[black] = E8;
  //   board->pmaterial[white] = board->pmaterial[black] =
  //   2*ValueR + 2*ValueN + 2*ValueB + ValueQ;
  // board->material[white] = board->material[black] =
  //    board->pmaterial[white] + 8*ValueP;
  //   CalcHashKey (board);

  //   UpdateFriends (board);
  // UpdateCBoard (board);
  // UpdateMvboard (board);

  // for (i = A1; i <= H8; i++)
  // {
  //    if (board->cboard[i])
  //    {
  //       SETBIT (board->blockerr90, r90[i]);
  //       SETBIT (board->blockerr45, r45[i]);
  //       SETBIT (board->blockerr315, r315[i]);
  //    }
  // }
  // board->castled[white] = board->castled[black] = false;
  // board->phase = PHASE(board);
  // board->Game50 = 0;
}

void PickStateInit(pickstate_t *pickstate){
  pickstate->movelist->gencnt=0;
  pickstate->captures->gencnt=0;
  pickstate->p=pickstate->movelist->moves;
  pickstate->phase=PICKHASH;
}


//void SearchParamsInit(searchparams_t *searchparams){
//   searchparams->NPS=0;
   //   searchparams->SearchTime = 5;
   //   searchparams->SearchDepth = 0;
//   searchparams->ST=0;
//   searchparams->TCMove=40;
//   searchparams->TCTime=5.0;
   //   searchparams->MoveLimit[0]=searchparams->TCMove;
   // searchparams->MoveLimit[1]=searchparams->TCMove;
//   searchparams->TimeLimit[0]=60*searchparams->TCTime;
//   searchparams->TimeLimit[1]=60*searchparams->TCTime;
//  searchparams->NextPeriod=2*searchparams->TCMove;
//}

void SearchStatsInit(searchstats_t *searchstats){
   searchstats->NodeCnt = searchstats->QuiesCnt = 0;
   searchstats->EvalCnt = searchstats->EvalCall = 0;
   searchstats->OneRepCnt = searchstats->ChkExtCnt = searchstats->RcpExtCnt = 
     searchstats->PawnExtCnt = searchstats->HorzExtCnt = 
     searchstats->ThrtExtCnt = 0;
   searchstats->KingExtCnt = 0;
   searchstats->NullCutCnt = searchstats->FutlCutCnt = 0;
   // [VdB]
   searchstats->LMRRedCnt=0;
   searchstats->GenCnt=0;
}

void SearchDataInit(searchdata_t *searchdata, board_t *board){
   memset(searchdata,0,sizeof(searchdata_t));
   //searchdata->RepeatCnt = 0;
   searchdata->RootPawns = nbits (board->b[white][pawn] | board->b[black][pawn]);
   searchdata->RootPieces = nbits (board->friends[white] | board->friends[black]) -
		searchdata->RootPawns;
   searchdata->RootMaterial = MATERIAL(board,board->side);
}








