/* GNU Chess 5.0 - search.c - tree-search code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
   bug-gnu-chess@gnu.org
   cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "eval.h"
#include "bitboard.h"
#include "move.h"
#include "movelist.h"
#include "option.h"
#include "timer.h"
#include "ttable.h"
#include "search.h"
#include "moveseq.h"

#define TIMECHECK 	0xFFF       // 4095
#define HISTSCORE(d)	((d)*(d))

#define FUTSCORE(board,fdel)     (MATERIAL(board,side)+(fdel))

search_t search;

void SearchInit(option_list_t *options){
    search.UseNull=OptionGetBool(options,"NullMovePruning");
    search.UseNullVer=OptionGetBool(options,"NullMoveVerification");
    search.UseTT=OptionGetBool(options,"UseTT");
    search.UseTTinPV=OptionGetBool(options,"UseTTinPV");
    search.UseMateScoresinPV=OptionGetBool(options,"UseMateScoresinPV");
    search.UseLMR=OptionGetBool(options,"LMR");
    search.UseFut=OptionGetBool(options,"FutilityPruning");
    search.UseAspiration=OptionGetBool(options,"AspirationSearch");
    search.UseMateDistancePruning=OptionGetBool(options,"MateDistancePruning");
}

static int GETNEXTMOVE(board_t *board, 
		       searchdata_t *searchdata, 
		       pickstate_t *pickstate,
		       int ply, 
		       leaf **p){
    int ret;
    //    ret=((searchdata)->InChk[ply] ? 
    ret=(board->in_check ? 
	 PhasePick1 (board, 
		     searchdata, 
		     pickstate,
		     p, 
		     ply): 
	 PhasePick (board, 
		    searchdata, 
		    pickstate,
		    p, 
		    ply));
    ASSERT(!ret || MoveIsPseudoLegal(board,(*p)->move));
    return ret;
}


void MoveMakeEx(board_t *board, 
		searchdata_t *searchdata, 
		int ply, 
		int *move, 
		undo_t *undo){
    searchdata->moveseq->moves[ply]=*move;
    searchdata->moveseq->gencnt=ply+1;
    MoveMakeS(board,move,undo);
}

void MoveUnMakeEx(board_t *board, 
		searchdata_t *searchdata, 
		undo_t *undo){
    searchdata->moveseq->gencnt--;
    MoveUnMakeS(board,undo);
}

static void NullMoveMakeEx(board_t *board, 
			   searchdata_t *searchdata, 
			   int ply, 
			   undo_t *undo){
    searchdata->moveseq->moves[ply]=NULLMOVE;
    NullMoveMakeS(board,undo);
}

static void NullMoveUnMakeEx(board_t *board, 
			   searchdata_t *searchdata, 
			   undo_t *undo){
    searchdata->moveseq->gencnt--;
    NullMoveUnMakeS(board,undo);
}


// [VdB] Periodic updates.

static void ShowUpdate(searchparams_t *searchparams, 
		       searchdata_t *searchdata,
		       searchstats_t *searchstats)
{
    Output("stat01: %d %ld %d %d %d %s\n", 
	   (int)(100*TimerElapsed (searchdata->timer)),
	   searchstats->NodeCnt+searchstats->QuiesCnt,
	   searchdata->RootDepth,
	   searchdata->RootMovesRemaining,
	   searchdata->RootMoves,
	   searchdata->RootMoveCurrent
	   );
}


// [VdB] Look at input during searching (not pondering)

void inline SearchParseInput(searchparams_t *searchparams, 
			     searchdata_t *searchdata, 
			     searchstats_t *searchstats){
    split_input();
    if(false){
	/*    }else if (flags & TIMEOUT){
	      longjmp(searchdata->rootpointer,1);*/
    }else if (flags & PONDER && !(flags & ANALYZE &&  tokeneq(token[0],"."))){
	longjmp(searchdata->rootpointer,1);
    }else if(tokeneq(token[0],"?")||
	     tokeneq(token[0],"force")||
	     tokeneq(token[0],"stop")){   // Some commands end the search.
	                                  // We include "stop" in case
	                                  // we want to implement the
	                                  // uci protocol in the future.
	longjmp(searchdata->rootpointer,1);
    }else if(tokeneq(token[0],"result")||
	     tokeneq(token[0],"new")||
	     tokeneq(token[0],"quit")){    // Some end the game as well.
	// In that case we don't return a 
	// move since it is useless.

	// In case of QUIT perhaps set flag!

	longjmp(searchdata->rootpointer,1);
    }else{ // We also process some command during search
	if(tokeneq(token[0],"post")){
	    cmd_post();
	}else if(tokeneq(token[0],"nopost")){
	    cmd_nopost();
	}else if(tokeneq(token[0],"easy")){
	    cmd_easy();
	}else if(tokeneq(token[0],"hard")){
	    cmd_hard();
	}else if(tokeneq(token[0],"ping")){
	    cmd_ping();
	}else if(tokeneq(token[0],"isready")){
	    cmd_isready();
	}else if(tokeneq(token[0],"option")){
	    cmd_option();
	}else if (flags & ANALYZE && tokeneq(token[0],".")){ /* periodic updates */
	    ShowUpdate(searchparams,searchdata,searchstats);
	}
	// For unrecognized commands or commands that have been 
	// processed here we discard the input.
	input_wakeup();
    }
}

void SearchCheckTime(searchparams_t *searchparams,
		     searchdata_t *searchdata,
		     searchstats_t *searchstats){
    searchstats->ElapsedTime = TimerElapsed(searchdata->timer);
    if(searchstats->ElapsedTime >= searchparams->maxtime){ /* Hard timeout */
	longjmp(searchdata->rootpointer,1);	
    }
    if(searchstats->ElapsedTime >= searchparams->SearchTime  && 
       searchdata->RootScore>searchdata->RootAlpha && searchdata->RootState!=SCORE_FAIL_LOW
       ){ /* Do not timeout while failing low, or while resolving a fail low.
	     Currently this also means we cannot timeout
	     while searching the first root move. Is this a good idea?
	  */
	longjmp(searchdata->rootpointer,1);	
    }
}


static void print_hash_score(board_t *board, int depth, int ply){
    int move, score;
    uint8_t ret;
    ret=TTableGet(TTable,board, depth,ply,&score,&move);
    switch(ret){
    case 0:
	Log("NOTFOUND ");
	break;
    case TTABLE_QUIESCENT:
	Log("QUIESCENT ");
	break;
    case TTABLE_LOWERBOUND:
	Log("LOWERBOUND ");
	break;
    case TTABLE_UPPERBOUND:
	Log("UPPERBOUND ");
	break;
    case TTABLE_EXACTSCORE:
	Log("EXACTSCORE ");
	break;
    case TTABLE_POORDRAFT:
	Log("POORDRAFT ");
	break;
    default:
	Log("UNKNOWN ");
    }
    if(ret){
	Log("move=%s score=%d",AlgbrMove(move),score);
    }
    Log("\n");
}


int SearchPVS(board_t *board, 
	      searchparams_t *searchparams, 
	      searchdata_t *searchdata, 
	      searchstats_t *searchstats,
	      uint8_t ply, 
	      short depth, 
	      int alpha, 
	      int beta, 
	      short nodetype, 
	      short parentnodetype,
	      moveseq_t *pv)
/* Search at a node with open window which is expected to fail high. */
{
    int score,savebeta;

    ASSERT(parentnodetype==PV);

    savebeta=beta;

    /* First search with zero window around beta hoping to 
       get a cutoff. 
    */
    score = Search (board, 
		    searchparams, 
		    searchdata, 
		    searchstats,
		    ply, 
		    depth, 
		    beta-1, 
		    beta, 
		    CUT, /* nodetype */
		    parentnodetype,
		    pv);
    ASSERT(MoveSeqValidate(board,pv));
    if (alpha < score && score < beta) { 
	/* We did not fail high and also not low...
	   We now we see if our score is strictly smaller than
	   the above returned upper bound.
	*/
	beta=score+1;  // +1 to get exact tt entries.
	score = Search (board, 
			searchparams, 
			searchdata, 
			searchstats,
			ply, 
			depth, 
			alpha, 
			beta, 
			PV, /* nodetype */
			parentnodetype,
			pv);
			    
	ASSERT(MoveSeqValidate(board,pv));
	    
	
    }
    if(score >=beta+1 && score<savebeta){
	/* Hmm, we have a case of search instability here... 
	   Research with fully open window.
	*/
	beta=savebeta;
	score = Search (board, 
			searchparams, 
			searchdata, 
			searchstats,
			ply, 
			depth, 
			alpha, 
			beta, 
			PV, /* nodetype */
			parentnodetype,
			pv);

	ASSERT(MoveSeqValidate(board,pv));

    }
    return score;
    
}

void SearchRoot (board_t *board, 
		 searchparams_t *searchparams, 
		 searchdata_t *searchdata,
		 searchstats_t *searchstats,
		 movelist_t *movelist)

/**************************************************************************
 *
 *  This perform searches at ply=1.  For ply>1, it calls the more generic
 *  search() routine. 
 *
 *  This routine returns both a score (searchdata->RootScore) which is reset
 *  at every invokation, and a move (searchstats->RootPV) which is only
 *  set if searchdata->RootScore > searchdata->RootAlpha.
 *
 **************************************************************************/


{
    int score, alpha,beta,savealpha;
    short depth;
    uint8_t side, xside;
    leaf *p;
    undo_t undo[1];
    char SANmv[SANSZ];
    moveseq_t pv[1];
    //    pv->gencnt=0;


    beta=searchdata->RootBeta;
    alpha=searchdata->RootAlpha;
    depth=searchdata->RootDepth;

    searchdata->RootScore=-INFINITY-1;
    savealpha = alpha;

    

    searchstats->NodeCnt++;
    
    side = board->side;
    xside = 1^side;

    searchdata->ChkCnt[1]=0;
    searchdata->ChkCnt[2]=0;
    searchdata->InChk[1]=board->in_check;
    if(board->in_check){
	searchstats->ChkExtCnt++;
	searchdata->ChkCnt[2]=1;
	depth += 1;
    }

    for (p = movelist->moves; p < movelist->moves+movelist->gencnt; p++) {
	MoveListPickSel(movelist,p); 
	
	/* periodic updates */
	MoveToSAN (board, p->move, searchdata->RootMoveCurrent);
	searchdata->RootMoves=movelist->gencnt;
	searchdata->RootMovesRemaining=movelist->moves+movelist->gencnt-p-1;
	/* end periodic updates */
	
	MoveMakeEx(board,searchdata,1,&p->move,undo);
	
	/*  If first move, search against full alpha-beta window  */
		
	if (p == movelist->moves){
	    
	    score = -Search (board, 
			     searchparams, 
			     searchdata,
			     searchstats,
			     2, 
			     depth-1, 
			     -beta, 
			     -alpha, 
			     PV, /* nodetype */
			     PV,
			     pv);

	    ASSERT(MoveSeqValidate(board,pv));

	} else { /*  we expect to fail low at other nodes */
	    alpha = MAX (searchdata->RootScore, alpha);  
	    score=-SearchPVS(board, 
			     searchparams, 
			     searchdata, 
			     searchstats,
			     2, 
			     depth-1, 
			     -beta, 
			     -alpha, 
			     PV, 
			     PV,
			     pv);

	}
	
	// XXX
	searchdata->moveseq->gencnt--;
	MoveUnMakeEx (board, searchdata, undo);
	
	p->score = score;  /* this is for sorting the movelist at the root */
	
	if (score > searchdata->RootScore) {
	    /* Since RootScore is initialized as -INFINITY-1
	       this branch will be taken at least once
	       (if there is no timeout).
	    */
	    searchdata->RootScore = score;  
	    if (searchdata->RootScore > alpha){  /* new PV */
		searchstats->RootPV = p->move;
		searchstats->RootPVScore=searchdata->RootScore;
		if (searchdata->RootScore >= beta){  /* fail high at root */

		    ShowLine (board, 
			      searchparams, 
			      searchdata, 
			      searchstats,
			      SCORE_FAIL_HIGH);
		    goto done;
		}
		MoveSeqPrependMove(searchdata->RootPVSeq,pv,searchstats->RootPV);		

		ShowLine (board, 
			  searchparams, 
			  searchdata, 
			  searchstats,
			  SCORE_EXACT);
	    }
	}
	
       
	if(input_status==INPUT_AVAILABLE){
	    SearchParseInput(searchparams,searchdata,searchstats);
	}
       
	if(!(flags & PONDER) && (searchstats->NodeCnt & TIMECHECK)==0){
	    SearchCheckTime(searchparams,searchdata,searchstats);
	}
	

    }
    
    
   

    /* We always get here if there is not timeout.
       This happens either by searching all rootmoves,
       or by a having a move with a score above beta.
    */

 done:

    if (searchdata->RootScore <= savealpha){

	/* Fail low at root */
	ShowLine (board, 
		  searchparams, 
		  searchdata, 
		  searchstats,
		  SCORE_FAIL_LOW);

	/*  If none of the moves is good, we still want to try the same first move */
	movelist->moves->score = savealpha;

	
    }else{   /* searchdata->RootScore>savealpha */

	/* no fail low: update history */
	searchdata->history[side][searchstats->RootPV & 0x0FFF] += HISTSCORE(depth);

        /* this is useful for analysis */
        if (search.UseTT){
            TTablePut (TTable,board, depth, 1, savealpha, beta, searchdata->RootScore, searchstats->RootPV);
        }

    }
    
    
    
}




// [VdB] Should be called after MakeMove (with side and xside like
// before make move.
#define WasPassedPawn(move) ((board->cboard[TOSQ((move))] == pawn &&	\
			      (board->b[xside][pawn] &			\
			       bitboards.PassedPawnMask[side][TOSQ((move))]) == NULLBITBOARD))






int Search (board_t *board, 
	    searchparams_t *searchparams, 
	    searchdata_t *searchdata, 
	    searchstats_t *searchstats,
	    uint8_t ply, 
	    short depth, 
	    int alpha, 
	    int beta, 
	    short nodetype, 
	    short parentnodetype,
	    moveseq_t *pv_out)
/**************************************************************************
 *
 *  The basic algorithm for this search routine came from Anthony 
 *  Marsland.  It is a PVS (Principal Variation Search) algorithm.
 *  The fail-soft alpha-beta technique is also used for improved
 *  pruning.
 *
 **************************************************************************/
{
    int best, score, nullscore, savealpha, savedepth;
    int side, xside;
    int rc, t0, t1, firstmove;
    int fcut, fdel, donull, savenode, extend;
    leaf *p, *pbest;
    int g0, g1;
    int unexpected_fail_low;
    int upperbound;
    // [VdB] LMR
    int movecount;
    pickstate_t pickstate[1];
    undo_t undo[1];
    int i;
    int R;
    int ret;
    moveseq_t pv[1];
    int hashcut;


    ASSERT((nodetype==PV && alpha+1<beta)|| (nodetype!=PV && alpha+1>=beta));
    
    pv->gencnt=0;
    pv_out->gencnt=0;

    savedepth=depth;
    savealpha = alpha;
    searchstats->NodeCnt++;


    // emergency exit

    if(ply>=MAXPLYDEPTH){
	best=board->in_check?alpha:Evaluate(board, 
					    searchparams,
					    searchdata, 
					    searchstats, 
					    alpha,
					    beta);
	goto nomove;
	
    }



    PickStateInit(pickstate);


    // XXX Move this up!
    if (EvaluateDraw (board)){
	// No move!
	best=DRAWSCORE;
	goto nomove;
    }

    if (board->GameCnt >= board->Game50+3 && BoardRepeat(board)) {
	// No move!
	best=DRAWSCORE;
	goto nomove;
    }

    if (board->GameCnt >= board->Game50+100) {
	// No move!
	best=DRAWSCORE;
	goto nomove;
    }


    /* End draw check */


    side = board->side;
    xside = 1^side;
    donull = true;

    /*************************************************************************
     *
     *  Perform some basic search extensions.
     *  1.  One reply extensions.  
     *  2.  If in check, extend (maximum of RootDepth-1).
     *  3.  If there is a threat to the King, extend (not beyond 2*RootDepth)
     *  4.  If recapture to same square and not beyond RootDepth+2
     *  5.  If pawn move to 7th rank at the leaf node, extend.
     *
     *************************************************************************/
    extend = false;
  
    // was only ply<=2
    //    searchdata->InChk[ply] = SqAtakd (board,board->king[side], xside);
    searchdata->InChk[ply]=board->in_check;



    //    if (searchdata->InChk[ply]) {
    if (board->in_check) {
	GenPseudoCheckEscapes (board, pickstate->movelist);
	if(pickstate->movelist->gencnt==0){
	    // No move!
	    best=-MATE+ply-2;
	    goto nomove;
	    
	}
	if (pickstate->movelist->gencnt==1){
	    depth += 1;
	    extend = true;
	    searchstats->OneRepCnt++;
	}
    }


    g0=searchdata->moveseq->moves[ply-1];
    g1=ply>2?searchdata->moveseq->moves[ply-2]:0;
    t0 = TOSQ(g0); 
    t1 = TOSQ(g1);
   
    searchdata->ChkCnt[ply+1] = searchdata->ChkCnt[ply];
   
    //    if (!extend && searchdata->InChk[ply]  && ply <= 2*searchdata->RootDepth){
    if (!extend && board->in_check  && ply <= 2*searchdata->RootDepth){
	searchstats->ChkExtCnt++;
	searchdata->ChkCnt[ply+1]++;
	depth += 1;
	extend = true;
    }
    // [VdB] Experiment: added nodetype==PV. Does not work as is.
    // We need the node type of the previous move....
    else if (!searchdata->KingThrt[side][ply-1] && 
	     searchdata->KingThrt[side][ply] && 
	     ply <= 2*searchdata->RootDepth)
	{
	    searchstats->KingExtCnt++;
	    extend = true;
	    depth += 1;
	    extend = true;
	    donull = false;
	}
    /* Promotion extension */
    else if (parentnodetype==PV && 
	     (g0 & PROMOTION))
	{
	    searchstats->PawnExtCnt++;
	    depth += 1; /* Not reached, but why?! */
	    extend = true;
	}
    /* Recapture extension */

    else if (parentnodetype==PV && 
	     (g0 & CAPTURE) && (board->material[searchparams->computer] - 
				board->material[1^searchparams->computer] == searchdata->RootMaterial))
	{
	    searchstats->RcpExtCnt++;
	    depth += 1;
	    extend = true;
	}
    /* 6th or 7th rank extension */
    else if (parentnodetype==PV && 
	     depth <= 1 && board->cboard[t0] == pawn && (RANK(t0) == rank7[xside] || RANK(t0) == rank6[xside]))
	{
	    searchstats->PawnExtCnt++;
	    depth += 1;
	    extend = true;
	}

    /**************************************************************************** 
     *
     *  The following extension is to handle cases when the opposing side is 
     *  delaying the mate by useless interposing moves. 
     *
     ****************************************************************************/
    if (ply > 2 && searchdata->InChk[ply-1] && board->cboard[t0] != king && t0 != t1 && 
	!SqAtakd (board,t0, xside))
	{
	    searchstats->HorzExtCnt++;
	    depth += 1;
	    extend = true;
	}

    /***************************************************************************
     *
     *  This is a new code to perform search reductiion.  We introduce some
     *  form of selectivity here.
     *
     **************************************************************************/


    if (depth <= 0){
	searchstats->NodeCnt--;
	best=Quiesce (board, searchparams, searchdata, searchstats, ply, alpha, beta); 
	goto nomove;
    }



    /**************************************************************************** 
     *
     *  Probe the transposition table for a score and a move.
     *  If the score is an upperbound, then we can use it to improve the value
     *  of beta.  If a lowerbound, we improve alpha.  If it is an exact score,
     *  if we now get a cut-off due to the new alpha/beta, return the score.
     *
     ***************************************************************************/


    searchdata->Hashmv[ply] = 0;
    upperbound = INFINITY;
    if (search.UseTT){ 
	    rc = TTableGet (TTable, board,  depth, ply, &score, &g1);
	    if(rc){
		searchdata->Hashmv[ply] = g1 & MOVEMASK;
	    }
	    hashcut=false;
	    if ((search.UseTTinPV 
		 || (search.UseMateScoresinPV && MATESCORE(score))
		||nodetype!=PV) 
		&& rc){
			switch (rc)
			    {
			    case TTABLE_POORDRAFT  :  /* Not reached */ 
				break;
			    case TTABLE_EXACTSCORE :  /* Not reached */ 
				best=score;
				hashcut=true;
				goto nomove;
				break;
			    case TTABLE_UPPERBOUND :  
				beta = MIN (beta, score);
				upperbound = score;
				donull = false;
				break;
			    case TTABLE_LOWERBOUND :  
				alpha = MAX (alpha, score);
				/* alpha = score;*/
				break;
			    case TTABLE_QUIESCENT  :  
				searchdata->Hashmv[ply] = 0;
				break;
			    default : break;
			    }
			if (alpha >= beta){
			    best=score;
			    hashcut=true;
			    goto nomove;
			}
		}
	    //}
	}




    /*****************************************************************************
     *
     *  Perform the null move here.  There are certain cases when null move
     *  is not done.  
     *  1.  When the previous move is a null move.
     *  2.  At the frontier (depth == 1)
     *  3.  At a PV node.
     *  4.  If side to move is in check.
     *  5.  If the material score + pawn value is still below beta.
     *  6.  If we are being mated at next ply.
     *  7.  If hash table indicate the real score is below beta (UPPERBOUND).
     *  8.  If side to move has less than or equal to a bishop in value.
     *  9.  If RootDepth <= 3.  This allows us to find mate-in 2 problems quickly.
     *  10. We are looking for a null threat.
     *
     *****************************************************************************/


    // stockfish!
    R=depth>=5?4:3;
    if(MATERIAL(board,side)>beta+2*ValueP){
        R++;
    }
    //    R=3;
    if (search.UseNull && g0 != NULLMOVE && depth > 1 && nodetype != PV &&
	!board->in_check && MATERIAL(board,side)+ValueP > beta && beta > -MATE+ply && donull &&
	board->pmaterial[side] > ValueB ){
	    NullMoveMakeEx(board,searchdata,ply,undo);
	    nullscore = -Search (board, 
				 searchparams, 
				 searchdata, 
				 searchstats,
				 ply+1, 
				 depth-R, 
				 -beta, 
				 -beta+1, 
				 ALL, 
				 nodetype,
				 pv);
	    //      UnmakeNullMove (board,xside); 
	    ASSERT(MoveSeqValidate(board,pv));

	    NullMoveUnMakeEx(board,searchdata,undo);
	    if (nullscore >= beta){
		if(depth<6 || !search.UseNullVer){
		    searchstats->NullCutCnt++;
		    best=nullscore /* beta*/;
		    goto nomove;
		}else{ /* verification */
		    int nullscore_ver;
		    int old_flags;
		    old_flags=search.UseNull; 
		    search.UseNull=false; // temporary hach
		    nullscore_ver=Search (board, 
					  searchparams, 
					  searchdata, 
					  searchstats,
					  ply, 
					  depth-5,  
					  beta-1, 
					  beta, 
					  nodetype, 
					  parentnodetype,
					  pv);
		    ASSERT(MoveSeqValidate(board,pv));

		    search.UseNull=old_flags;
		    if(nullscore_ver>=beta){
			searchstats->NullCutCnt++;
			best=nullscore /* beta */;
			goto nomove; 
		    }
		}
	    }
	    if ( depth-3 >= 1 && MATERIAL(board,side) > beta && nullscore <= -MATE+256){
		depth += 1;
		extend = true;
	    }
	}

    if (board->in_check && 1<pickstate->movelist->gencnt){
	SortMoves (board,searchdata,pickstate->movelist,ply);
    }

    GETNEXTMOVE(board,searchdata,pickstate,ply,&p);


    /*************************************************************************
     *
     *  Razoring + Futility.
     *  At depth 3, if there is no extensions and we are really bad, decrease
     *  the search depth by 1.
     *  At depth 2, if there is no extensions and we are quite bad, then we
     *  prune all non checking moves and capturing moves that don't bring us up
     *  back to alpha.
     *  Caveat: Skip all this if we are in the ending.
     *
     *************************************************************************/
    fcut = false;
    fdel = MAX (ValueQ, searchdata->maxposnscore[side]);


    if (!extend && nodetype != PV && depth == 3 && FUTSCORE(board,fdel) <= alpha)
	{
	    depth = 2;
	    searchstats->RazrCutCnt++;
	}
    fdel = MAX (ValueR, searchdata->maxposnscore[side]);
    fcut = (!extend && nodetype != PV && depth == 2 && FUTSCORE(board,fdel) <= alpha);
    if (!fcut)
	{
	    fdel = MAX (3*ValueP, searchdata->maxposnscore[side]);
	    fcut = (nodetype != PV && depth == 1 && FUTSCORE(board,fdel) <= alpha);
	}


    // Debug futility pruning

    //   fcut=false;


    fcut=fcut && search.UseFut;

    MoveMakeEx (board, searchdata, ply,  &p->move, undo);
    g0 = g1 = 0;
    while ((g0 = board->xin_check) > 0 ||
	   (fcut && FUTSCORE(board,fdel) < alpha && 
	    !board->in_check
	    // new XXX
	    //	    && pickstate->phase==PICKREST
	    	    	    && !(WasPassedPawn(p->move))
	    //&& !(board->cboard[TOSQ(p->move)] == pawn && (RANK(TOSQ(p->move)) == rank7[side] || RANK(TOSQ(p->move)) == rank6[side]))
	    // end new XXX
	    ))
	{
	    if (g0 == 0) g1++;
	    MoveUnMakeEx(board,searchdata,undo);
	    if (GETNEXTMOVE(board,searchdata,pickstate,ply,&p) == false){
		      best= (g1 ? Evaluate(board, 
						   searchparams,
						   searchdata, 
						   searchstats, 
						   alpha,
						   beta) 
				     /* MATERIAL(board,side) */ /* some alternatives */
				     /* FUTSCORE(board,fdel) */
				     /* alpha */
				     : DRAWSCORE);
		      goto nomove;
	    }
	    MoveMakeEx (board, searchdata, ply, &p->move, undo);
    }      

    firstmove = true;
    pbest = p;
    best = -INFINITY;
    //    savealpha = alpha;
    nullscore = INFINITY;
    savenode = nodetype;
    if (nodetype != PV){
	nodetype = (nodetype == CUT) ? ALL : CUT;
    }
    // [VdB] LMR
    movecount=1;
    while (1) {
	/* We have already made the move before the loop. */
	if (firstmove){
	    firstmove = false;
	    unexpected_fail_low=false;
	    score = -Search (board, 
			     searchparams, 
			     searchdata, 
			     searchstats,
			     ply+1, 
			     depth-1, 
			     -beta, 
			     -alpha, 
			     nodetype, 
			     savenode,
			     pv);

	    ASSERT(MoveSeqValidate(board,pv));

	    if(score<=alpha && (savenode==CUT|| savenode==PV)){
		unexpected_fail_low=true;
	    }
	    unexpected_fail_low=false;
	} else {  /* Zero window search for rest of moves */
	    if (GETNEXTMOVE(board,searchdata,pickstate,ply,&p) == false)
		break;
	    // [VdB] LMR
	    movecount++;
	    MoveMakeEx (board, searchdata, ply, &p->move, undo);

	    //	    	    if (SqAtakd (board,board->king[side], xside)) 
			if (board->xin_check) 
		{
		    MoveUnMakeEx(board,searchdata,undo);
		    continue;
		}

	    //	    searchdata->InChk[ply+1]=SqAtakd (board,board->king[xside], side);

	    /*****************************************************************************
	     *
	     *  Futility pruning.  The idea is that at the frontier node (depth == 1),
	     *  if the side on the move is materially bad, then if the move doesn't win
	     *  back material or the move isn't a check or doesn't threatened the king, 
	     *  then there is no point in searching this move.  So skip it.  
	     *  Caveat:  However if the node is a PV, we skip this test.
	     *
	     *****************************************************************************/
	    //	    if (fcut && FUTSCORE(board,fdel) <= alpha && !searchdata->InChk[ply+1] 
	    if (fcut && FUTSCORE(board,fdel) <= alpha && !board->in_check
		// new XXX
		//		&& pickstate->phase==PICKREST
				 && !(WasPassedPawn(p->move))
		//&& !(board->cboard[TOSQ(p->move)] == pawn && (RANK(TOSQ(p->move)) == rank7[side] || RANK(TOSQ(p->move)) == rank6[side]))
		// end new XXX
		/*&&!MateScan (board,xside)*/
		)
		
		{
		    //	   UnmakeMove (board, xside, &p->move);
		    MoveUnMakeEx(board,searchdata,undo);
		    searchstats->FutlCutCnt++;

		    continue;
		}

	    if (nodetype == PV)
		nodetype = CUT;
	    alpha = MAX (best, alpha);                /* fail-soft condition */
	    // [VdB] LMR
	    if(search.UseLMR && 
	       depth>=2 && 
	       pickstate->phase==PICKREST && 
	       //	       /*!extend  &&*/ !searchdata->InChk[ply+1] && 
	       !board->in_check &&
	       (unexpected_fail_low || (savenode!=PV && movecount>=3) || movecount>=6)
	       && !(WasPassedPawn(p->move))){
		score = -Search (board, 
				 searchparams, 
				 searchdata, 
				 searchstats,
				 ply+1, 
				 depth-2, 
				 -alpha-1, 
				 -alpha, 
				 nodetype, 
				 savenode,
				 pv);	
		ASSERT(MoveSeqValidate(board,pv));

	    }else{
		score=alpha+1;// to trigger next condition
	    }
	    searchstats->LMRRedCnt++;
	    if(score>alpha){//research
		searchstats->LMRRedCnt--;
		score = -Search (board, 
				 searchparams, 
				 searchdata, 
				 searchstats,
				 ply+1, 
				 depth-1, 
				 -alpha-1, 
				 -alpha, 
				 nodetype, 
				 savenode,
				 pv);
		ASSERT(MoveSeqValidate(board,pv));
    
	    }
	    // End LMR
	    if (score > best){
		if (savenode == PV){
			nodetype = PV;
		}
		if (alpha < score && score < beta){
		    int tempalpha;
		    //			    tempalpha=score;
		    tempalpha=alpha; 
		    score = -Search (board, 
				     searchparams, 
				     searchdata, 
				     searchstats,
				     ply+1, 
				     depth-1, 
				     -beta, 
				     -tempalpha, 
				     nodetype, 
				     savenode,
				     pv);
		    ASSERT(MoveSeqValidate(board,pv));
		    
		} 
		
	    }
	}

	//      UnmakeMove (board, xside, &p->move);
	MoveUnMakeEx(board,searchdata,undo);

	if (score > best){
	  best = score;
	  pbest = p;
	  if (best >= beta){
	    goto done;
	  }
	  if(best>alpha){
	      MoveSeqPrependMove(pv_out,pv,p->move);
	  }
	}

	if(input_status==INPUT_AVAILABLE){
	    SearchParseInput(searchparams,searchdata,searchstats);
	}

	if(!(flags & PONDER) && (searchstats->NodeCnt & TIMECHECK)==0){
	    SearchCheckTime(searchparams,searchdata,searchstats);
	}


	/*  The following line should be explained as I occasionally forget too :) */
	/*  This code means that if at this ply, a mating move has been found,     */
	/*  then we can skip the rest of the moves!  				   */
	if (search.UseMateDistancePruning && (MATE+1 == best+ply))
		    goto done;
    } 

    /*****************************************************************************
     *
     *  Out of main search loop.
     *
     *****************************************************************************/
 done:


    
    /*
      if (upperbound < best)
      printf ("Inconsistencies %d %d\n", upperbound, best);
    */

    /*  Save the best move inside the transposition table  */
    if (search.UseTT){
	TTablePut (TTable, board, depth, ply, savealpha, beta, best, pbest->move); 
    }

    /*  Update history  */
    if (best > savealpha){
	searchdata->history[side][pbest->move & 0x0FFF] += HISTSCORE(depth);
    }

    /*  Don't store captures as killers as they are tried before killers */
    if (!(pbest->move & (CAPTURE | PROMOTION)) && best > savealpha)
	{
	    if (searchdata->killer1[ply] == 0)
		searchdata->killer1[ply] = pbest->move & MOVEMASK;
	    else if ((pbest->move & MOVEMASK) != searchdata->killer1[ply])
		searchdata->killer2[ply] = pbest->move & MOVEMASK;
	}

 nomove:

       ASSERT(hashcut|| MATESCORE(best)|| best==DRAWSCORE || best<=savealpha || best>=beta || beta<=savealpha+1 || pv_out->gencnt>=savedepth);

    return (best);
}

void ShowLine (board_t *board, 
	       searchparams_t *searchparams, 
	       searchdata_t *searchdata, 
	       searchstats_t *searchstats,
	       int c)
/*****************************************************************************
 *
 *  Print out the latest PV found during the search.
 *
 *****************************************************************************/
{
    int i, score, post_score;
    int pvar;
    char SANmv[SANSZ];
    moveseq_t moveseq[1];
    char pv[MAXPLYDEPTH*(SANSZ+1)+1];
    board_t board_work[1];
    char *move2;



    if (!(flags & POST)){
	return;
    }

    searchstats->ElapsedTime = TimerElapsed(searchdata->timer);
    MoveToSAN (board, searchstats->RootPV, SANmv);

    if(c==SCORE_PV){
	score=(int)searchstats->RootPVScore;
    }else{
	score=(int)searchdata->RootScore;
    }

    post_score=score;
    if(flags & PONDER && !(flags & ANALYZE)) post_score=-post_score;

    Output ("%d %d %d %ld ", 
	    searchdata->RootDepth, 
	    post_score, 
	    (int)(100*searchstats->ElapsedTime), 
	    searchstats->NodeCnt+searchstats->QuiesCnt);
	 

    if (c == SCORE_FAIL_LOW){
	Output (" :-(\n");
	return;
    } else if (c == SCORE_FAIL_HIGH){

	if(flags&PONDER && !(flags & ANALYZE)){
	    Output (" (%s!)\n", SANmv);
	}else{
	    Output (" %s!\n", SANmv);
	}
	return;
    }

    if(flags&PONDER && !(flags & ANALYZE)){
	Output (" (%s)", SANmv);
    }else{
	Output(" %s", SANmv);
    }
    


    MoveSeqToSAN(board,searchdata->RootPVSeq,pv);
    move2=strchr(pv,' ');
    if(move2){
	Output("%s",move2);
    }

    Output ("\n");
}

