/* GNU Chess 5.0 - ttable.c - transposition table code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
     cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
*/

#include <stdio.h>
#include <string.h>
#include "common.h"
#include "moveseq.h"
#include "ttable.h"

ttable_t TTable[1];

int ClusterSize=4;

void TTableInit(ttable_t *ttable){
    ttable->Slot=NULL;
    ttable->Size=0;
    ttable->Mask=0;
    ttable->Age=0;
    TTableResetStats(ttable);    
}

void TTableZero (ttable_t *ttable)
/****************************************************************************
 *   
 *  Zero out the transposition table.
 *
 ****************************************************************************/
{
   memset (ttable->Slot, 0, ttable->Size * sizeof (ttable_slot_t));
}



void TTableClear(ttable_t *ttable){
    if(ttable->Slot!=NULL){
	free(ttable->Slot);
    }
    TTableInit(ttable);
}

void TTableResetStats(ttable_t *ttable){
    ttable->TotalGetHashCnt=0;
    ttable->GoodGetHashCnt=0;
    ttable->TotalPutHashCnt=0;
    ttable->CollHashCnt=0;
}

void TTableIncAge(ttable_t *ttable){
    ttable->Age++;
}

void TTableSetSizeMb(ttable_t *ttable, unsigned int size){
    int slotsize;
    slotsize=(1048576*size)/(sizeof(ttable_slot_t));
    TTableSetSize(ttable,slotsize);
}

void TTableSetSize(ttable_t *ttable, unsigned int size)
/****************************************************************************
  *
  *  Allocate memory for our transposition tables.  By default, the
  *  transposition table will have 16K entries; 8K for each color.
  *  Add code to allocated memory for pawn table as well.
  *
  ****************************************************************************/
{
    unsigned int tsize;
    unsigned int i;

    TTableClear(ttable);
    i = (size < HASHSLOTS ? HASHSLOTS : size);
        
    ttable->Mask = 0;
    while ((i>>=1) > 0)	{
	ttable->Mask <<= 1;
	ttable->Mask |= 1;
    }
    ttable->Size = ttable->Mask + 1+(ClusterSize-1);
    ASSERT(size/2<ttable->Size<=size+(ClusterSize-1));

    OutputConsole("Adjusting HashSize to %d slots\n", ttable->Size);  
   
    ttable->Slot = malloc (ttable->Size * sizeof (ttable_slot_t));
    if (ttable->Slot == NULL) {
         OutputConsole("Not enough memory for transposition table,\n");
	 exit(EXIT_FAILURE);
    }
    tsize = (ttable->Size *  sizeof (ttable_slot_t)) >> 10;
    OutputConsole("Transposition table:  Entries=%dK Size=%dK\n",
              ttable->Size>>10, tsize);
}



void TTablePut (ttable_t *ttable,
		board_t *board, 
		uint8_t depth, 
		uint8_t ply, 
		int alpha, 
		int beta, 
		int score, 
		int move){

    uint32_t quality; // the higher the worse
    uint32_t best; 
    ttable_slot_t *best_slot, *t, *s;
    int i;

    ASSERT(move!=0);
    ASSERT(MoveIsPseudoLegal(board,move));

    t = ttable->Slot + (board->HashKey & ttable->Mask); 
    best=0;
    // best_slot=t;

    for(s=t;s<=t+ClusterSize-1;s++){
	if(s->key==board->HashKey){
	    best_slot=s;
	    break;
	}
	quality=((ttable->Age-s->age)<<16)+(255-s->depth)<<4+(7-s->flag);
	if(quality>best){
	    best=quality;
	    best_slot=s;
	    
	}
    }
    if (best_slot->flag){
      ttable->CollHashCnt++;
    }
    ttable->TotalPutHashCnt++;
    best_slot->move = move;
    best_slot->key = board->HashKey;
    best_slot->depth = depth;
    best_slot->age=ttable->Age;
    if (best_slot->depth == 0){
	best_slot->flag = TTABLE_QUIESCENT;
    }else if (score >= beta){
	best_slot->flag = TTABLE_LOWERBOUND;         
    }else if (score <= alpha){
	best_slot->flag = TTABLE_UPPERBOUND;
    }else{
	best_slot->flag = TTABLE_EXACTSCORE;
    }
    if (MATESCORE(score))
	best_slot->score = score + ( score > 0 ? ply : -ply);
    else
	best_slot->score = score;
    
    ASSERT(best_slot->score>=-MATE-1);
    ASSERT(best_slot->score<=MATE+1);

}



uint8_t TTableGet(ttable_t *ttable,
		  board_t *board, 
		  uint8_t depth, 
		  uint8_t ply, 
		  int *score, 
		     int *move){
    ttable_slot_t *s,*t;
    int i;
    t = ttable->Slot + (board->HashKey & ttable->Mask); 
    for(s=t;s<=t+ClusterSize-1;s++){
	if(s->key==board->HashKey){
	    ttable->GoodGetHashCnt++;
	    *move = s->move;
	    *score = s->score;
	    if (s->depth == 0){
		return (TTABLE_QUIESCENT);
	    }
	    if (s->depth < depth && !(MATESCORE (s->score))){
		return (TTABLE_POORDRAFT);
	    }
	    if (MATESCORE(*score)){
		*score -= (*score > 0 ? ply : -ply);
	    }
	    return s->flag;
	}
    }
    return 0;
}


void TTableGetPV(ttable_t *ttable,
		 board_t *board, 
		 int score,
		 int maxgen,
		 moveseq_t *moveseq
		  ){
    ttable_slot_t *s,*t;
    int i,j;
    int found;
    board_t board_work[1];
    moveseq->gencnt=0;
    memcpy(board_work,board,sizeof(board_t));
    while(1){
	if ((MATESCORE(score) && abs(score) == MATE-moveseq->gencnt) || 
	    BoardRepeat (board_work) || moveseq->gencnt==maxgen){
	    break;
	}
	t = ttable->Slot + (board_work->HashKey & ttable->Mask); 
	found=false;
	for(s=t;s<=t+ClusterSize-1;s++){
	    if(s->key==board_work->HashKey && s->flag==TTABLE_EXACTSCORE && 
		(!(moveseq->gencnt & 1 && score == s->score)||
		 ((moveseq->gencnt & 1) && score == -s->score))
	       ){
		ASSERT(s->move!=0);
		moveseq->moves[moveseq->gencnt]=s->move;
		found=true;
		break;
	    }
	}
	if(!found){
	    break;
	}else{
	    MoveMake(board_work,moveseq->moves[moveseq->gencnt]);
	    (moveseq->gencnt)++;
	}
    }
}

void TTableFeedPV(ttable_t *ttable,
		  board_t *board,
		  uint8_t depth,
		  int score,
		  moveseq_t *moveseq){
    int i;
    board_t board_work[1];
    memcpy(board_work,board,sizeof(board_t));
    for(i=0;i<moveseq->gencnt;i++){
	TTablePut(ttable,
		  board_work,
		  depth-i,                      /*depth*/
		  i+1,                          /*ply*/
		  -INFINITY,                    /*alpha*/
		  INFINITY,                     /*beta*/
		  i%2?score:-score,             /*score*/
		  moveseq->moves[i]             /*move*/
		  );
	MoveMake(board_work,moveseq->moves[i]);
    }
}
    



void TTShow(ttable_t *ttable, board_t *board){
    ttable_slot_t *s, *t;
    int move, score;
    uint8_t ret;
    uint8_t stored_depth;
    uint16_t age;
    Output("Age=%u Hash=%llx \n",ttable->Age, board->HashKey);
    t = ttable->Slot + (board->HashKey & ttable->Mask); 

    for(s=t;s<=t+ClusterSize-1;s++){
	if(s->key==board->HashKey){
	    move = s->move;
	    score = s->score;
	    age=s->age;
	    stored_depth=s->depth;
	    
	    
	    switch(s->flag){
	    case 0:
		Output("NOTFOUND ");
		break;
	    case TTABLE_QUIESCENT:
		Output("QUIESCENT ");
		break;
	    case TTABLE_LOWERBOUND:
		Output("LOWERBOUND ");
		break;
	    case TTABLE_UPPERBOUND:
		Output("UPPERBOUND ");
		break;
	    case TTABLE_EXACTSCORE:
		Output("EXACTSCORE ");
		break;
	    case TTABLE_POORDRAFT:
		Output("POORDRAFT ");
		break;
	    default:
		Output("UNKNOWN ");
	    }
	    if(s->flag){
		Output("move=%s score=%d depth=%d age=%u\n",AlgbrMove(move),score,stored_depth, age);
	    }
	}
    }
    Output("\n");
}


void PTClear (void)
/****************************************************************************
 *   
 *  Zero out the pawn transposition table.
 *
 ****************************************************************************/
{
   memset (PawnTab[white], 0, PAWNSLOTS * sizeof (PawnSlot));
   memset (PawnTab[black], 0, PAWNSLOTS * sizeof (PawnSlot));
}


