/* GNU Chess 5.0 - util.c - utility routine code
   Copyright (c) 1999-2002 Free Software Foundation, Inc.

   GNU Chess is based on the two research programs 
   Cobalt by Chua Kong-Sian and Gazebo by Stuart Cracraft.

   GNU Chess is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   GNU Chess is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Chess; see the file COPYING.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.

   Contact Info: 
     bug-gnu-chess@gnu.org
     cracraft@ai.mit.edu, cracraft@stanfordalumni.org, cracraft@earthlink.net
*/
/*
 * Endianness checks are now unnecessary
 */

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>

#include "common.h"
#include "bitboard.h"
#include "option.h"
#include "eval.h"
#include "log.h"

// [VdB] Some things borrowed from Polyglot (but simplified)

void my_string_set(const char * * variable, const char string[]) {

   ASSERT(variable!=NULL);
   ASSERT(string!=NULL);

   if (*variable != NULL) free((void*)(*variable));
   *variable = strdup(string);
}

int my_string_case_equal(const char string_1[], const char string_2[]) {

   int c1, c2;

   ASSERT(string_1!=NULL);
   ASSERT(string_2!=NULL);

   while (true) {

      c1 = *string_1++;
      c2 = *string_2++;

      if (tolower(c1) != tolower(c2)) return false;
      if (c1 == '\0') return true;
   }

   return true;
}

int my_string_equal(const char string_1[], const char string_2[]) {

   ASSERT(string_1!=NULL);
   ASSERT(string_2!=NULL);

   return strcmp(string_1,string_2) == 0;
}

void my_string_clear(const char * * variable) {

   ASSERT(variable!=NULL);

   if (*variable != NULL) {
      free((void*)(*variable));
      *variable = NULL;
   }
}

void my_string_append(const char **dst, const char *src){
  char *s;
  if(*dst){
    //XXX
    s=malloc(strlen(*dst)+strlen(src)+1);
    sprintf(s,"%s",*dst);
    strcat(s,src);
    free((void *)(*dst));
    *dst=s;
  }else{
    my_string_set(dst,src);
  }
}


void my_path_join(char *join_path, const char *path, const char *file){
    char separator;
#ifdef _WIN32
    separator='\\';
#else
    separator='/';
#endif
    snprintf(join_path,MAXSTR,"%s%c%s",path,separator,file);
    join_path[MAXSTR-1]='\0';
}

void my_sleep(int msec){
#ifndef _WIN32
  struct timespec tm;
  tm.tv_sec=msec/1000;
  tm.tv_nsec=1000000*(msec%1000);
  nanosleep(&tm,NULL);
#else
  Sleep(msec);
#endif
}


void Output(char  *fmt, ...){
  va_list ap;
  char buf[MAXSTR];
  va_start(ap, fmt);
  vsnprintf(buf, MAXSTR-1, fmt, ap);
  va_end(ap);
  printf("%s",buf);
  fflush(stdout);
  LogLock(LogDefault);
  LogSetNewlinePrefix(LogDefault,"OUT: ");
  Log("%s",buf);
  LogSetNewlinePrefix(LogDefault,DEFAULT_NEWLINE_PREFIX);
  LogUnlock(LogDefault);
}

void OutputConsole(char  *fmt, ...){
  va_list ap;
  char buf[MAXSTR];
  va_start(ap, fmt);
  vsnprintf(buf, MAXSTR-1, fmt, ap);
  va_end(ap);
  if(Protocol->console){
    printf("%s",buf);
    fflush(stdout);
  }
  LogLock(LogDefault);
  LogSetNewlinePrefix(LogDefault,"OUT: ");
  Log("%s",buf);
  LogSetNewlinePrefix(LogDefault,DEFAULT_NEWLINE_PREFIX);
  LogUnlock(LogDefault);
}

/*
 * We actually do not have this configuration variable yet,
 * should be manually added when inlining is not possible.
 * If inlining is possible, these functions have now moved
 * to util.h which must be included by all callers of
 * leadz() and nbits().
 */

#ifdef NO_INLINE



unsigned char leadz (BitBoard b)
/**************************************************************************
 *
 *  Returns the leading bit in a BitBoard.  Leftmost bit is 0 and
 *  rightmost bit is 63.  Thanks to Robert Hyatt for this algorithm.
 *
 ***************************************************************************/
{
  if (b >> 48) return lzArray[b >> 48];
  if (b >> 32) return lzArray[b >> 32] + 16;
  if (b >> 16) return lzArray[b >> 16] + 32;
  return lzArray[b] + 48;
}

unsigned char nbits (BitBoard b)
/***************************************************************************
 *
 *  Count the number of bits in b.
 *
 ***************************************************************************/
{
  return BitCount[b>>48] + BitCount[(b>>32) & 0xffff]
    + BitCount[(b>>16) & 0xffff] + BitCount[b & 0xffff];
}

#endif /* NO_INLINE */

void UpdateBlockers(board_t *board){
  BitBoard b;
  int sq;
  b=board->blocker;
  board->blockerr90=0;
  board->blockerr45=0;
  board->blockerr315=0;
  while(b){
    sq=leadz(b);
    CLEARBIT(b,sq);
    SETBIT (board->blockerr90, r90[sq]);
    SETBIT (board->blockerr45, r45[sq]);
    SETBIT (board->blockerr315, r315[sq]);

  }
}

void UpdateFriends (board_t *board)
/***************************************************************************
 *
 *  Update friend and enemy bitboard.
 *
 ***************************************************************************/
{
   register BitBoard *w, *b;

   w = board->b[white];
   b = board->b[black];
   board->friends[white] = 
      w[pawn] | w[knight] | w[bishop] | w[rook] | w[queen] | w[king];
   board->friends[black] = 
      b[pawn] | b[knight] | b[bishop] | b[rook] | b[queen] | b[king];
   board->blocker = board->friends[white] | board->friends[black];
}
   

void UpdateCBoard (board_t *board)
/**************************************************************************
 *
 *  Updates cboard[].  cboard[i] returns the piece on square i.
 *
 **************************************************************************/
{
   BitBoard b;
   int piece, sq;

   memset (board->cboard, 0, sizeof (board->cboard));
   for (piece = pawn; piece <= king; piece++)
   {
      b = board->b[white][piece] | board->b[black][piece];
      while (b)
      {
         sq = leadz (b);
         CLEARBIT (b, sq);
         board->cboard[sq] = piece;
      }
   }
}


static const int OrigCboard[64] = 
{ rook,  knight, bishop, queen, king,  bishop, knight, rook,
  pawn,  pawn,   pawn,   pawn,  pawn,  pawn,   pawn,   pawn,
  empty, empty,  empty,  empty, empty, empty,  empty,  empty,
  empty, empty,  empty,  empty, empty, empty,  empty,  empty,
  empty, empty,  empty,  empty, empty, empty,  empty,  empty,
  empty, empty,  empty,  empty, empty, empty,  empty,  empty,
  pawn,  pawn,   pawn,   pawn,  pawn,  pawn,   pawn,   pawn,
  rook,  knight, bishop, queen, king,  bishop, knight, rook };

void UpdateMvboard (board_t *board)
/**************************************************************************
 *
 *  Updates Mvboard[].  Mvboard[i] returns the number of times the piece
 *  on square i has moved.  When loading from EPD, if a piece is not on
 *  its original square, set it to 1, otherwise set to 0.
 *
 **************************************************************************/
{
   int sq;
 
   for (sq = 0; sq < 64; sq++)
   {
      if (board->cboard[sq] == empty || board->cboard[sq] == OrigCboard[sq])
         board->Mvboard[sq] = 0;
      else
         board->Mvboard[sq] = 1;
   } 
}


void UpdatePhase(board_t *board){
    int total_material=board->material[white]+board->material[black];
  
    while(total_material>Phase[board->phase]){
	board->phase--;
    }
    while(total_material<=Phase[board->phase+1]){
	board->phase++;
    }

    ASSERT(board->material[white]+board->material[black]<=Phase[board->phase]);
    ASSERT(board->material[white]+board->material[black]>Phase[board->phase+1]);
  
}





